# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from datetime import datetime, date, time
import ipaddress
from uuid import uuid1, uuid4

from cassandra.util import OrderedMap, Date, Time, sortedset, Duration

from tests.integration import get_server_versions


PRIMITIVE_DATATYPES = sortedset([
    'ascii',
    'bigint',
    'blob',
    'boolean',
    'decimal',
    'double',
    'float',
    'inet',
    'int',
    'text',
    'timestamp',
    'timeuuid',
    'uuid',
    'varchar',
    'varint',
])

PRIMITIVE_DATATYPES_KEYS = PRIMITIVE_DATATYPES.copy()

COLLECTION_TYPES = sortedset([
    'list',
    'set',
    'map',
])


def update_datatypes():
    _cass_version, _cql_version = get_server_versions()

    if _cass_version >= (2, 1, 0):
        COLLECTION_TYPES.add('tuple')

    if _cass_version >= (2, 2, 0):
        PRIMITIVE_DATATYPES.update(['date', 'time', 'smallint', 'tinyint'])
        PRIMITIVE_DATATYPES_KEYS.update(['date', 'time', 'smallint', 'tinyint'])
    if _cass_version >= (3, 10):
        PRIMITIVE_DATATYPES.add('duration')

    global SAMPLE_DATA
    SAMPLE_DATA = get_sample_data()


def get_sample_data():
    sample_data = {}

    for datatype in PRIMITIVE_DATATYPES:
        if datatype == 'ascii':
            sample_data[datatype] = 'ascii'

        elif datatype == 'bigint':
            sample_data[datatype] = 2 ** 63 - 1

        elif datatype == 'blob':
            sample_data[datatype] = bytearray(b'hello world')

        elif datatype == 'boolean':
            sample_data[datatype] = True

        elif datatype == 'decimal':
            sample_data[datatype] = Decimal('12.3E+7')

        elif datatype == 'double':
            sample_data[datatype] = 1.23E+8

        elif datatype == 'float':
            sample_data[datatype] = 3.4028234663852886e+38

        elif datatype == 'inet':
            sample_data[datatype] = ('123.123.123.123',
                                     '2001:db8:85a3:8d3:1319:8a2e:370:7348',
                                     ipaddress.IPv4Address("123.123.123.123"),
                                     ipaddress.IPv6Address('2001:db8:85a3:8d3:1319:8a2e:370:7348'))

        elif datatype == 'int':
            sample_data[datatype] = 2147483647

        elif datatype == 'text':
            sample_data[datatype] = 'text'

        elif datatype == 'timestamp':
            sample_data[datatype] = datetime(2013, 12, 31, 23, 59, 59, 999000)

        elif datatype == 'timeuuid':
            sample_data[datatype] = uuid1()

        elif datatype == 'uuid':
            sample_data[datatype] = uuid4()

        elif datatype == 'varchar':
            sample_data[datatype] = 'varchar'

        elif datatype == 'varint':
            sample_data[datatype] = int(str(2147483647) + '000')

        elif datatype == 'date':
            sample_data[datatype] = Date(date(2015, 1, 15))

        elif datatype == 'time':
            sample_data[datatype] = Time(time(16, 47, 25, 7))

        elif datatype == 'tinyint':
            sample_data[datatype] = 123

        elif datatype == 'smallint':
            sample_data[datatype] = 32523

        elif datatype == 'duration':
            sample_data[datatype] = Duration(months=2, days=12, nanoseconds=21231)

        else:
            raise Exception("Missing handling of {0}".format(datatype))

    return sample_data

SAMPLE_DATA = get_sample_data()


def get_sample(datatype):
    """
    Helper method to access created sample data for primitive types
    """
    if isinstance(SAMPLE_DATA[datatype], tuple):
        return SAMPLE_DATA[datatype][0]
    return SAMPLE_DATA[datatype]


def get_all_samples(datatype):
    """
    Helper method to access created sample data for primitive types
    """
    if isinstance(SAMPLE_DATA[datatype], tuple):
        return SAMPLE_DATA[datatype]
    return SAMPLE_DATA[datatype],


def get_collection_sample(collection_type, datatype):
    """
    Helper method to access created sample data for collection types
    """

    if collection_type == 'list':
        return [get_sample(datatype), get_sample(datatype)]
    elif collection_type == 'set':
        return sortedset([get_sample(datatype)])
    elif collection_type == 'map':
        return OrderedMap([(get_sample(datatype), get_sample(datatype))])
    elif collection_type == 'tuple':
        return (get_sample(datatype),)
    else:
        raise Exception('Missing handling of non-primitive type {0}.'.format(collection_type))



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from tests.integration import CCM_KWARGS, use_cluster, remove_cluster, MockLoggingHandler
from tests.integration import setup_keyspace

from cassandra.cluster import Cluster
from cassandra import cluster

from collections import namedtuple
from functools import wraps
import logging
from threading import Thread, Event
from ccmlib.node import TimeoutError
import time
import logging

import unittest


def setup_module():
    remove_cluster()


UPGRADE_CLUSTER_NAME = "upgrade_cluster"
UpgradePath = namedtuple('UpgradePath', ('name', 'starting_version', 'upgrade_version', 'configuration_options'))

log = logging.getLogger(__name__)


class upgrade_paths(object):
    """
    Decorator used to specify the upgrade paths for a particular method
    """
    def __init__(self, paths):
        self.paths = paths

    def __call__(self, method):
        @wraps(method)
        def wrapper(*args, **kwargs):
            for path in self.paths:
                self_from_decorated = args[0]
                log.debug('setting up {path}'.format(path=path))
                self_from_decorated.UPGRADE_PATH = path
                self_from_decorated._upgrade_step_setup()
                method(*args, **kwargs)
                log.debug('tearing down {path}'.format(path=path))
                self_from_decorated._upgrade_step_teardown()
        return wrapper


class UpgradeBase(unittest.TestCase):
    """
    Base class for the upgrade tests. The _setup method
    will clean the environment and start the appropriate C* version according
    to the upgrade path. The upgrade can be done in a different thread using the
    start_upgrade upgrade_method (this would be the most realistic scenario)
    or node by node, waiting for the upgrade to happen, using _upgrade_one_node method
    """
    UPGRADE_PATH = None
    start_cluster = True
    set_keyspace = True

    @classmethod
    def setUpClass(cls):
        cls.logger_handler = MockLoggingHandler()
        logger = logging.getLogger(cluster.__name__)
        logger.addHandler(cls.logger_handler)

    def _upgrade_step_setup(self):
        """
        This is not the regular _setUp method because it will be called from
        the decorator instead of letting nose handle it.
        This setup method will start a cluster with the right version according
        to the variable UPGRADE_PATH.
        """
        remove_cluster()
        self.cluster = use_cluster(UPGRADE_CLUSTER_NAME + self.UPGRADE_PATH.name, [3],
                                   ccm_options=self.UPGRADE_PATH.starting_version, set_keyspace=self.set_keyspace,
                                   configuration_options=self.UPGRADE_PATH.configuration_options)
        self.nodes = self.cluster.nodelist()
        self.last_node_upgraded = None
        self.upgrade_done = Event()
        self.upgrade_thread = None

        if self.start_cluster:
            setup_keyspace()

            self.cluster_driver = Cluster()
            self.session = self.cluster_driver.connect()
            self.logger_handler.reset()

    def _upgrade_step_teardown(self):
        """
        special tearDown method called by the decorator after the method has ended
        """
        if self.upgrade_thread:
            self.upgrade_thread.join(timeout=5)
            self.upgrade_thread = None

        if self.start_cluster:
            self.cluster_driver.shutdown()

    def start_upgrade(self, time_node_upgrade):
        """
        Starts the upgrade in a different thread
        """
        log.debug('Starting upgrade in new thread')
        self.upgrade_thread = Thread(target=self._upgrade, args=(time_node_upgrade,))
        self.upgrade_thread.start()

    def _upgrade(self, time_node_upgrade):
        """
        Starts the upgrade in the same thread
        """
        start_time = time.time()
        for node in self.nodes:
            self.upgrade_node(node)
            end_time = time.time()
            time_to_upgrade = end_time - start_time
            if time_node_upgrade > time_to_upgrade:
                time.sleep(time_node_upgrade - time_to_upgrade)
        self.upgrade_done.set()

    def is_upgraded(self):
        """
        Returns True if the upgrade has finished and False otherwise
        """
        return self.upgrade_done.is_set()

    def wait_for_upgrade(self, timeout=None):
        """
        Waits until the upgrade has completed
        """
        self.upgrade_done.wait(timeout=timeout)

    def upgrade_node(self, node):
        """
        Upgrades only one node. Return True if the upgrade
        has finished and False otherwise
        """
        node.drain()
        node.stop(gently=True)

        node.set_install_dir(**self.UPGRADE_PATH.upgrade_version)

        # There must be a cleaner way of doing this, but it's necessary here
        # to call the private method from cluster __update_topology_files
        self.cluster._Cluster__update_topology_files()
        try:
            node.start(wait_for_binary_proto=True, wait_other_notice=True)
        except TimeoutError:
            self.fail("Error starting C* node while upgrading")

        return True


class UpgradeBaseAuth(UpgradeBase):
    """
    Base class of authentication test, the authentication parameters for
    C* still have to be specified within the upgrade path variable
    """
    start_cluster = False
    set_keyspace = False


    def _upgrade_step_setup(self):
        """
        We sleep here for the same reason as we do in test_authentication.py:
        there seems to be some race, with some versions of C* taking longer to
        get the auth (and default user) setup. Sleep here to give it a chance
        """
        super(UpgradeBaseAuth, self)._upgrade_step_setup()
        time.sleep(10)



import pytest

from tests.integration.cloud import setup_package, teardown_package

@pytest.fixture(scope='session', autouse=True)
def setup_and_teardown_packages():
    setup_package()
    yield
    teardown_package()


# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from cassandra.protocol import ProtocolHandler, ResultMessage, QueryMessage, UUIDType, read_int
from cassandra.query import tuple_factory, SimpleStatement
from cassandra.cluster import (ResponseFuture, ExecutionProfile, EXEC_PROFILE_DEFAULT,
    ContinuousPagingOptions, NoHostAvailable)
from cassandra import ProtocolVersion, ConsistencyLevel

from tests.integration import use_singledc, drop_keyspace_shutdown_cluster, \
    greaterthanorequalcass30, execute_with_long_wait_retry, greaterthanorequaldse51, greaterthanorequalcass3_10, \
    TestCluster, greaterthanorequalcass40, requirecassandra
from tests.integration.datatype_utils import update_datatypes, PRIMITIVE_DATATYPES
from tests.integration.standard.utils import create_table_with_all_types, get_all_primitive_params

import uuid
from unittest import mock


def setup_module():
    use_singledc()
    update_datatypes()


class CustomProtocolHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster = TestCluster()
        cls.session = cls.cluster.connect()
        cls.session.execute("CREATE KEYSPACE custserdes WITH replication = { 'class' : 'SimpleStrategy', 'replication_factor': '1'}")
        cls.session.set_keyspace("custserdes")

    @classmethod
    def tearDownClass(cls):
        drop_keyspace_shutdown_cluster("custserdes", cls.session, cls.cluster)

    def test_custom_raw_uuid_row_results(self):
        """
        Test to validate that custom protocol handlers work with raw row results

        Connect and validate that the normal protocol handler is used.
        Re-Connect and validate that the custom protocol handler is used.
        Re-Connect and validate that the normal protocol handler is used.

        @since 2.7
        @jira_ticket PYTHON-313
        @expected_result custom protocol handler is invoked appropriately.

        @test_category data_types:serialization
        """

        # Ensure that we get normal uuid back first
        cluster = TestCluster(
            execution_profiles={EXEC_PROFILE_DEFAULT: ExecutionProfile(row_factory=tuple_factory)}
        )
        session = cluster.connect(keyspace="custserdes")

        result = session.execute("SELECT schema_version FROM system.local")
        uuid_type = result[0][0]
        self.assertEqual(type(uuid_type), uuid.UUID)

        # use our custom protocol handlder
        session.client_protocol_handler = CustomTestRawRowType
        result_set = session.execute("SELECT schema_version FROM system.local")
        raw_value = result_set[0][0]
        self.assertTrue(isinstance(raw_value, bytes))
        self.assertEqual(len(raw_value), 16)

        # Ensure that we get normal uuid back when we re-connect
        session.client_protocol_handler = ProtocolHandler
        result_set = session.execute("SELECT schema_version FROM system.local")
        uuid_type = result_set[0][0]
        self.assertEqual(type(uuid_type), uuid.UUID)
        cluster.shutdown()

    def test_custom_raw_row_results_all_types(self):
        """
        Test to validate that custom protocol handlers work with varying types of
        results

        Connect, create a table with all sorts of data. Query the data, make the sure the custom results handler is
        used correctly.

        @since 2.7
        @jira_ticket PYTHON-313
        @expected_result custom protocol handler is invoked with various result types

        @test_category data_types:serialization
        """
        # Connect using a custom protocol handler that tracks the various types the result message is used with.
        cluster = TestCluster(
            execution_profiles={EXEC_PROFILE_DEFAULT: ExecutionProfile(row_factory=tuple_factory)}
        )
        session = cluster.connect(keyspace="custserdes")
        session.client_protocol_handler = CustomProtocolHandlerResultMessageTracked

        colnames = create_table_with_all_types("alltypes", session, 1)
        columns_string = ", ".join(colnames)

        # verify data
        params = get_all_primitive_params(0)
        results = session.execute("SELECT {0} FROM alltypes WHERE primkey=0".format(columns_string))[0]
        for expected, actual in zip(params, results):
            self.assertEqual(actual, expected)
        # Ensure we have covered the various primitive types
        self.assertEqual(len(CustomResultMessageTracked.checked_rev_row_set), len(PRIMITIVE_DATATYPES)-1)
        cluster.shutdown()

    @requirecassandra
    @greaterthanorequalcass40
    def test_protocol_divergence_v5_fail_by_continuous_paging(self):
        """
        Test to validate that V5 and DSE_V1 diverge. ContinuousPagingOptions is not supported by V5

        @since DSE 2.0b3 GRAPH 1.0b1
        @jira_ticket PYTHON-694
        @expected_result NoHostAvailable will be risen when the continuous_paging_options parameter is set

        @test_category connection
        """
        cluster = TestCluster(protocol_version=ProtocolVersion.V5, allow_beta_protocol_version=True)
        session = cluster.connect()

        max_pages = 4
        max_pages_per_second = 3
        continuous_paging_options = ContinuousPagingOptions(max_pages=max_pages,
                                                            max_pages_per_second=max_pages_per_second)

        future = self._send_query_message(session, timeout=session.default_timeout,
                                        consistency_level=ConsistencyLevel.ONE,
                            continuous_paging_options=continuous_paging_options)

        # This should raise NoHostAvailable because continuous paging is not supported under ProtocolVersion.DSE_V1
        with self.assertRaises(NoHostAvailable) as context:
            future.result()
        self.assertIn("Continuous paging may only be used with protocol version ProtocolVersion.DSE_V1 or higher",
                    str(context.exception))

        cluster.shutdown()

    @greaterthanorequalcass30
    def test_protocol_divergence_v4_fail_by_flag_uses_int(self):
        """
        Test to validate that the _PAGE_SIZE_FLAG is not treated correctly in V4 if the flags are
        written using write_uint instead of write_int

        @since 3.9
        @jira_ticket PYTHON-713
        @expected_result the fetch_size=1 parameter will be ignored

        @test_category connection
        """
        self._protocol_divergence_fail_by_flag_uses_int(ProtocolVersion.V4, uses_int_query_flag=False,
                                                        int_flag=True)

    @requirecassandra
    @greaterthanorequalcass40
    def test_protocol_v5_uses_flag_int(self):
        """
        Test to validate that the _PAGE_SIZE_FLAG is treated correctly using write_uint for V5

        @jira_ticket PYTHON-694
        @expected_result the fetch_size=1 parameter will be honored

        @test_category connection
        """
        self._protocol_divergence_fail_by_flag_uses_int(ProtocolVersion.V5, uses_int_query_flag=True, beta=True,
                                                        int_flag=True)

    @greaterthanorequaldse51
    def test_protocol_dsev1_uses_flag_int(self):
        """
        Test to validate that the _PAGE_SIZE_FLAG is treated correctly using write_uint for DSE_V1

        @jira_ticket PYTHON-694
        @expected_result the fetch_size=1 parameter will be honored

        @test_category connection
        """
        self._protocol_divergence_fail_by_flag_uses_int(ProtocolVersion.DSE_V1, uses_int_query_flag=True,
                                                        int_flag=True)

    @requirecassandra
    @greaterthanorequalcass40
    def test_protocol_divergence_v5_fail_by_flag_uses_int(self):
        """
        Test to validate that the _PAGE_SIZE_FLAG is treated correctly using write_uint for V5

        @jira_ticket PYTHON-694
        @expected_result the fetch_size=1 parameter will be honored

        @test_category connection
        """
        self._protocol_divergence_fail_by_flag_uses_int(ProtocolVersion.V5, uses_int_query_flag=False, beta=True,
                                                        int_flag=False)

    @greaterthanorequaldse51
    def test_protocol_divergence_dsev1_fail_by_flag_uses_int(self):
        """
        Test to validate that the _PAGE_SIZE_FLAG is treated correctly using write_uint for DSE_V1

        @jira_ticket PYTHON-694
        @expected_result the fetch_size=1 parameter will be honored

        @test_category connection
        """
        self._protocol_divergence_fail_by_flag_uses_int(ProtocolVersion.DSE_V1, uses_int_query_flag=False,
                                                        int_flag=False)

    def _send_query_message(self, session, timeout, **kwargs):
        query = "SELECT * FROM test3rf.test"
        message = QueryMessage(query=query, **kwargs)
        future = ResponseFuture(session, message, query=None, timeout=timeout)
        future.send_request()
        return future

    def _protocol_divergence_fail_by_flag_uses_int(self, version, uses_int_query_flag, int_flag = True, beta=False):
        cluster = TestCluster(protocol_version=version, allow_beta_protocol_version=beta)
        session = cluster.connect()

        query_one = SimpleStatement("INSERT INTO test3rf.test (k, v) VALUES (1, 1)")
        query_two = SimpleStatement("INSERT INTO test3rf.test (k, v) VALUES (2, 2)")

        execute_with_long_wait_retry(session, query_one)
        execute_with_long_wait_retry(session, query_two)

        with mock.patch('cassandra.protocol.ProtocolVersion.uses_int_query_flags', new=mock.Mock(return_value=int_flag)):
            future = self._send_query_message(session, 10,
                                              consistency_level=ConsistencyLevel.ONE, fetch_size=1)

            response = future.result()

            # This means the flag are not handled as they are meant by the server if uses_int=False
            self.assertEqual(response.has_more_pages, uses_int_query_flag)

        execute_with_long_wait_retry(session, SimpleStatement("TRUNCATE test3rf.test"))
        cluster.shutdown()


class CustomResultMessageRaw(ResultMessage):
    """
    This is a custom Result Message that is used to return raw results, rather then
    results which contain objects.
    """
    my_type_codes = ResultMessage.type_codes.copy()
    my_type_codes[0xc] = UUIDType
    type_codes = my_type_codes

    def recv_results_rows(self, f, protocol_version, user_type_map, result_metadata, column_encryption_policy):
            self.recv_results_metadata(f, user_type_map)
            column_metadata = self.column_metadata or result_metadata
            rowcount = read_int(f)
            self.parsed_rows = [self.recv_row(f, len(column_metadata)) for _ in range(rowcount)]
            self.column_names = [c[2] for c in column_metadata]
            self.column_types = [c[3] for c in column_metadata]


class CustomTestRawRowType(ProtocolHandler):
    """
    This is the a custom protocol handler that will substitute the the
    customResultMesageRowRaw Result message for our own implementation
    """
    my_opcodes = ProtocolHandler.message_types_by_opcode.copy()
    my_opcodes[CustomResultMessageRaw.opcode] = CustomResultMessageRaw
    message_types_by_opcode = my_opcodes


class CustomResultMessageTracked(ResultMessage):
    """
    This is a custom Result Message that is use to track what primitive types
    have been processed when it receives results
    """
    my_type_codes = ResultMessage.type_codes.copy()
    my_type_codes[0xc] = UUIDType
    type_codes = my_type_codes
    checked_rev_row_set = set()

    def recv_results_rows(self, f, protocol_version, user_type_map, result_metadata, column_encryption_policy):
        self.recv_results_metadata(f, user_type_map)
        column_metadata = self.column_metadata or result_metadata
        rowcount = read_int(f)
        rows = [self.recv_row(f, len(column_metadata)) for _ in range(rowcount)]
        self.column_names = [c[2] for c in column_metadata]
        self.column_types = [c[3] for c in column_metadata]
        self.checked_rev_row_set.update(self.column_types)
        self.parsed_rows = [
            tuple(ctype.from_binary(val, protocol_version)
                  for ctype, val in zip(self.column_types, row))
            for row in rows]


class CustomProtocolHandlerResultMessageTracked(ProtocolHandler):
    """
    This is the a custom protocol handler that will substitute the the
    CustomTestRawRowTypeTracked Result message for our own implementation
    """
    my_opcodes = ProtocolHandler.message_types_by_opcode.copy()
    my_opcodes[CustomResultMessageTracked.opcode] = CustomResultMessageTracked
    message_types_by_opcode = my_opcodes



"""
Helper module to populate a dummy Cassandra tables with data.
"""

from tests.integration.datatype_utils import PRIMITIVE_DATATYPES, get_sample


def create_table_with_all_types(table_name, session, N):
    """
    Method that given a table_name and session construct a table that contains
    all possible primitive types.

    :param table_name: Name of table to create
    :param session: session to use for table creation
    :param N: the number of items to insert into the table

    :return: a list of column names
    """
    # create table
    alpha_type_list = ["primkey int PRIMARY KEY"]
    col_names = ["primkey"]
    start_index = ord('a')
    for i, datatype in enumerate(PRIMITIVE_DATATYPES):
        alpha_type_list.append("{0} {1}".format(chr(start_index + i), datatype))
        col_names.append(chr(start_index + i))

    session.execute("CREATE TABLE {0} ({1})".format(
                        table_name, ', '.join(alpha_type_list)), timeout=120)

    # create the input

    for key in range(N):
        params = get_all_primitive_params(key)

        # insert into table as a simple statement
        columns_string = ', '.join(col_names)
        placeholders = ', '.join(["%s"] * len(col_names))
        session.execute("INSERT INTO {0} ({1}) VALUES ({2})".format(
                            table_name, columns_string, placeholders), params, timeout=120)
    return col_names


def get_all_primitive_params(key):
    """
    Simple utility method used to give back a list of all possible primitive data sample types.
    """
    params = [key]
    for datatype in PRIMITIVE_DATATYPES:
        # Also test for empty strings
        if key == 1 and datatype == 'ascii':
            params.append('')
        else:
            params.append(get_sample(datatype))
    return params


def get_primitive_datatypes():
    return ['int'] + list(PRIMITIVE_DATATYPES)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import unittest

from datetime import datetime
from uuid import uuid4
from packaging.version import Version
import uuid

from cassandra.cluster import Session
from cassandra import InvalidRequest
from tests.integration.cqlengine.base import BaseCassEngTestCase
from cassandra.cqlengine.connection import NOT_SET
from unittest import mock
from cassandra.cqlengine import functions
from cassandra.cqlengine.management import sync_table, drop_table
from cassandra.cqlengine.models import Model
from cassandra.cqlengine import columns
from cassandra.cqlengine import query
from cassandra.cqlengine.query import QueryException, BatchQuery
from datetime import timedelta
from datetime import tzinfo

from cassandra.cqlengine import statements
from cassandra.cqlengine import operators
from cassandra.util import uuid_from_time
from cassandra.cqlengine.connection import get_session
from tests.integration import PROTOCOL_VERSION, CASSANDRA_VERSION, greaterthancass20, greaterthancass21, \
    greaterthanorequalcass30, TestCluster
from tests.integration.cqlengine import execute_count, DEFAULT_KEYSPACE


class TzOffset(tzinfo):
    """Minimal implementation of a timezone offset to help testing with timezone
    aware datetimes.
    """

    def __init__(self, offset):
        self._offset = timedelta(hours=offset)

    def utcoffset(self, dt):
        return self._offset

    def tzname(self, dt):
        return 'TzOffset: {}'.format(self._offset.hours)

    def dst(self, dt):
        return timedelta(0)


class TestModel(Model):

    test_id = columns.Integer(primary_key=True)
    attempt_id = columns.Integer(primary_key=True)
    description = columns.Text()
    expected_result = columns.Integer()
    test_result = columns.Integer()


class IndexedTestModel(Model):

    test_id = columns.Integer(primary_key=True)
    attempt_id = columns.Integer(index=True)
    description = columns.Text()
    expected_result = columns.Integer()
    test_result = columns.Integer(index=True)


class CustomIndexedTestModel(Model):

    test_id = columns.Integer(primary_key=True)
    description = columns.Text(custom_index=True)
    indexed = columns.Text(index=True)
    data = columns.Text()


class IndexedCollectionsTestModel(Model):

    test_id = columns.Integer(primary_key=True)
    attempt_id = columns.Integer(index=True)
    description = columns.Text()
    expected_result = columns.Integer()
    test_result = columns.Integer(index=True)
    test_list = columns.List(columns.Integer, index=True)
    test_set = columns.Set(columns.Integer, index=True)
    test_map = columns.Map(columns.Text, columns.Integer, index=True)

    test_list_no_index = columns.List(columns.Integer, index=False)
    test_set_no_index = columns.Set(columns.Integer, index=False)
    test_map_no_index = columns.Map(columns.Text, columns.Integer, index=False)


class TestMultiClusteringModel(Model):

    one = columns.Integer(primary_key=True)
    two = columns.Integer(primary_key=True)
    three = columns.Integer(primary_key=True)


class TestQuerySetOperation(BaseCassEngTestCase):

    def test_query_filter_parsing(self):
        """
        Tests the queryset filter method parses it's kwargs properly
        """
        query1 = TestModel.objects(test_id=5)
        assert len(query1._where) == 1

        op = query1._where[0]

        assert isinstance(op, statements.WhereClause)
        assert isinstance(op.operator, operators.EqualsOperator)
        assert op.value == 5

        query2 = query1.filter(expected_result__gte=1)
        assert len(query2._where) == 2

        op = query2._where[1]
        self.assertIsInstance(op, statements.WhereClause)
        self.assertIsInstance(op.operator, operators.GreaterThanOrEqualOperator)
        assert op.value == 1

    def test_query_expression_parsing(self):
        """ Tests that query experessions are evaluated properly """
        query1 = TestModel.filter(TestModel.test_id == 5)
        assert len(query1._where) == 1

        op = query1._where[0]
        assert isinstance(op, statements.WhereClause)
        assert isinstance(op.operator, operators.EqualsOperator)
        assert op.value == 5

        query2 = query1.filter(TestModel.expected_result >= 1)
        assert len(query2._where) == 2

        op = query2._where[1]
        self.assertIsInstance(op, statements.WhereClause)
        self.assertIsInstance(op.operator, operators.GreaterThanOrEqualOperator)
        assert op.value == 1

    def test_using_invalid_column_names_in_filter_kwargs_raises_error(self):
        """
        Tests that using invalid or nonexistant column names for filter args raises an error
        """
        with self.assertRaises(query.QueryException):
            TestModel.objects(nonsense=5)

    def test_using_nonexistant_column_names_in_query_args_raises_error(self):
        """
        Tests that using invalid or nonexistant columns for query args raises an error
        """
        with self.assertRaises(AttributeError):
            TestModel.objects(TestModel.nonsense == 5)

    def test_using_non_query_operators_in_query_args_raises_error(self):
        """
        Tests that providing query args that are not query operator instances raises an error
        """
        with self.assertRaises(query.QueryException):
            TestModel.objects(5)

    def test_queryset_is_immutable(self):
        """
        Tests that calling a queryset function that changes it's state returns a new queryset
        """
        query1 = TestModel.objects(test_id=5)
        assert len(query1._where) == 1

        query2 = query1.filter(expected_result__gte=1)
        assert len(query2._where) == 2
        assert len(query1._where) == 1

    def test_queryset_limit_immutability(self):
        """
        Tests that calling a queryset function that changes it's state returns a new queryset with same limit
        """
        query1 = TestModel.objects(test_id=5).limit(1)
        assert query1._limit == 1

        query2 = query1.filter(expected_result__gte=1)
        assert query2._limit == 1

        query3 = query1.filter(expected_result__gte=1).limit(2)
        assert query1._limit == 1
        assert query3._limit == 2

    def test_the_all_method_duplicates_queryset(self):
        """
        Tests that calling all on a queryset with previously defined filters duplicates queryset
        """
        query1 = TestModel.objects(test_id=5)
        assert len(query1._where) == 1

        query2 = query1.filter(expected_result__gte=1)
        assert len(query2._where) == 2

        query3 = query2.all()
        assert query3 == query2

    def test_queryset_with_distinct(self):
        """
        Tests that calling distinct on a queryset w/without parameter are evaluated properly.
        """

        query1 = TestModel.objects.distinct()
        self.assertEqual(len(query1._distinct_fields), 1)

        query2 = TestModel.objects.distinct(['test_id'])
        self.assertEqual(len(query2._distinct_fields), 1)

        query3 = TestModel.objects.distinct(['test_id', 'attempt_id'])
        self.assertEqual(len(query3._distinct_fields), 2)

    def test_defining_only_fields(self):
        """
        Tests defining only fields

        @since 3.5
        @jira_ticket PYTHON-560
        @expected_result deferred fields should not be returned

        @test_category object_mapper
        """
        # simple only definition
        q = TestModel.objects.only(['attempt_id', 'description'])
        self.assertEqual(q._select_fields(), ['attempt_id', 'description'])

        with self.assertRaises(query.QueryException):
            TestModel.objects.only(['nonexistent_field'])

        # Cannot define more than once only fields
        with self.assertRaises(query.QueryException):
            TestModel.objects.only(['description']).only(['attempt_id'])

        # only with defer fields
        q = TestModel.objects.only(['attempt_id', 'description'])
        q = q.defer(['description'])
        self.assertEqual(q._select_fields(), ['attempt_id'])

        # Eliminate all results confirm exception is thrown
        q = TestModel.objects.only(['description'])
        q = q.defer(['description'])
        with self.assertRaises(query.QueryException):
            q._select_fields()

        q = TestModel.objects.filter(test_id=0).only(['test_id', 'attempt_id', 'description'])
        self.assertEqual(q._select_fields(), ['attempt_id', 'description'])

        # no fields to select
        with self.assertRaises(query.QueryException):
            q = TestModel.objects.only(['test_id']).defer(['test_id'])
            q._select_fields()

        with self.assertRaises(query.QueryException):
            q = TestModel.objects.filter(test_id=0).only(['test_id'])
            q._select_fields()

    def test_defining_defer_fields(self):
        """
        Tests defining defer fields

        @since 3.5
        @jira_ticket PYTHON-560
        @jira_ticket PYTHON-599
        @expected_result deferred fields should not be returned

        @test_category object_mapper
        """

        # simple defer definition
        q = TestModel.objects.defer(['attempt_id', 'description'])
        self.assertEqual(q._select_fields(), ['test_id', 'expected_result', 'test_result'])

        with self.assertRaises(query.QueryException):
            TestModel.objects.defer(['nonexistent_field'])

        # defer more than one
        q = TestModel.objects.defer(['attempt_id', 'description'])
        q = q.defer(['expected_result'])
        self.assertEqual(q._select_fields(), ['test_id', 'test_result'])

        # defer with only
        q = TestModel.objects.defer(['description', 'attempt_id'])
        q = q.only(['description', 'test_id'])
        self.assertEqual(q._select_fields(), ['test_id'])

        # Eliminate all results confirm exception is thrown
        q = TestModel.objects.defer(['description', 'attempt_id'])
        q = q.only(['description'])
        with self.assertRaises(query.QueryException):
            q._select_fields()

        # implicit defer
        q = TestModel.objects.filter(test_id=0)
        self.assertEqual(q._select_fields(), ['attempt_id', 'description', 'expected_result', 'test_result'])

        # when all fields are defered, it fallbacks select the partition keys
        q = TestModel.objects.defer(['test_id', 'attempt_id', 'description', 'expected_result', 'test_result'])
        self.assertEqual(q._select_fields(), ['test_id'])


class BaseQuerySetUsage(BaseCassEngTestCase):

    @classmethod
    def setUpClass(cls):
        super(BaseQuerySetUsage, cls).setUpClass()
        drop_table(TestModel)
        drop_table(IndexedTestModel)
        drop_table(CustomIndexedTestModel)

        sync_table(TestModel)
        sync_table(IndexedTestModel)
        sync_table(CustomIndexedTestModel)
        sync_table(TestMultiClusteringModel)

        TestModel.objects.create(test_id=0, attempt_id=0, description='try1', expected_result=5, test_result=30)
        TestModel.objects.create(test_id=0, attempt_id=1, description='try2', expected_result=10, test_result=30)
        TestModel.objects.create(test_id=0, attempt_id=2, description='try3', expected_result=15, test_result=30)
        TestModel.objects.create(test_id=0, attempt_id=3, description='try4', expected_result=20, test_result=25)

        TestModel.objects.create(test_id=1, attempt_id=0, description='try5', expected_result=5, test_result=25)
        TestModel.objects.create(test_id=1, attempt_id=1, description='try6', expected_result=10, test_result=25)
        TestModel.objects.create(test_id=1, attempt_id=2, description='try7', expected_result=15, test_result=25)
        TestModel.objects.create(test_id=1, attempt_id=3, description='try8', expected_result=20, test_result=20)

        TestModel.objects.create(test_id=2, attempt_id=0, description='try9', expected_result=50, test_result=40)
        TestModel.objects.create(test_id=2, attempt_id=1, description='try10', expected_result=60, test_result=40)
        TestModel.objects.create(test_id=2, attempt_id=2, description='try11', expected_result=70, test_result=45)
        TestModel.objects.create(test_id=2, attempt_id=3, description='try12', expected_result=75, test_result=45)

        IndexedTestModel.objects.create(test_id=0, attempt_id=0, description='try1', expected_result=5, test_result=30)
        IndexedTestModel.objects.create(test_id=1, attempt_id=1, description='try2', expected_result=10, test_result=30)
        IndexedTestModel.objects.create(test_id=2, attempt_id=2, description='try3', expected_result=15, test_result=30)
        IndexedTestModel.objects.create(test_id=3, attempt_id=3, description='try4', expected_result=20, test_result=25)

        IndexedTestModel.objects.create(test_id=4, attempt_id=0, description='try5', expected_result=5, test_result=25)
        IndexedTestModel.objects.create(test_id=5, attempt_id=1, description='try6', expected_result=10, test_result=25)
        IndexedTestModel.objects.create(test_id=6, attempt_id=2, description='try7', expected_result=15, test_result=25)
        IndexedTestModel.objects.create(test_id=7, attempt_id=3, description='try8', expected_result=20, test_result=20)

        IndexedTestModel.objects.create(test_id=8, attempt_id=0, description='try9', expected_result=50, test_result=40)
        IndexedTestModel.objects.create(test_id=9, attempt_id=1, description='try10', expected_result=60,
                                        test_result=40)
        IndexedTestModel.objects.create(test_id=10, attempt_id=2, description='try11', expected_result=70,
                                        test_result=45)
        IndexedTestModel.objects.create(test_id=11, attempt_id=3, description='try12', expected_result=75,
                                        test_result=45)

        if CASSANDRA_VERSION >= Version('2.1'):
            drop_table(IndexedCollectionsTestModel)
            sync_table(IndexedCollectionsTestModel)
            IndexedCollectionsTestModel.objects.create(test_id=12, attempt_id=3, description='list12', expected_result=75,
                                                       test_result=45, test_list=[1, 2, 42], test_set=set([1, 2, 3]),
                                                       test_map={'1': 1, '2': 2, '3': 3})
            IndexedCollectionsTestModel.objects.create(test_id=13, attempt_id=3, description='list13', expected_result=75,
                                                       test_result=45, test_list=[3, 4, 5], test_set=set([4, 5, 42]),
                                                       test_map={'1': 5, '2': 6, '3': 7})
            IndexedCollectionsTestModel.objects.create(test_id=14, attempt_id=3, description='list14', expected_result=75,
                                                       test_result=45, test_list=[1, 2, 3], test_set=set([1, 2, 3]),
                                                       test_map={'1': 1, '2': 2, '3': 42})

            IndexedCollectionsTestModel.objects.create(test_id=15, attempt_id=4, description='list14', expected_result=75,
                                                       test_result=45, test_list_no_index=[1, 2, 3], test_set_no_index=set([1, 2, 3]),
                                                       test_map_no_index={'1': 1, '2': 2, '3': 42})

    @classmethod
    def tearDownClass(cls):
        super(BaseQuerySetUsage, cls).tearDownClass()
        drop_table(TestModel)
        drop_table(IndexedTestModel)
        drop_table(CustomIndexedTestModel)
        drop_table(TestMultiClusteringModel)


class TestQuerySetCountSelectionAndIteration(BaseQuerySetUsage):

    @execute_count(2)
    def test_count(self):
        """ Tests that adding filtering statements affects the count query as expected """
        assert TestModel.objects.count() == 12

        q = TestModel.objects(test_id=0)
        assert q.count() == 4

    @execute_count(2)
    def test_query_expression_count(self):
        """ Tests that adding query statements affects the count query as expected """
        assert TestModel.objects.count() == 12

        q = TestModel.objects(TestModel.test_id == 0)
        assert q.count() == 4

    @execute_count(3)
    def test_iteration(self):
        """ Tests that iterating over a query set pulls back all of the expected results """
        q = TestModel.objects(test_id=0)
        # tuple of expected attempt_id, expected_result values
        compare_set = set([(0, 5), (1, 10), (2, 15), (3, 20)])
        for t in q:
            val = t.attempt_id, t.expected_result
            assert val in compare_set
            compare_set.remove(val)
        assert len(compare_set) == 0

        # test with regular filtering
        q = TestModel.objects(attempt_id=3).allow_filtering()
        assert len(q) == 3
        # tuple of expected test_id, expected_result values
        compare_set = set([(0, 20), (1, 20), (2, 75)])
        for t in q:
            val = t.test_id, t.expected_result
            assert val in compare_set
            compare_set.remove(val)
        assert len(compare_set) == 0

        # test with query method
        q = TestModel.objects(TestModel.attempt_id == 3).allow_filtering()
        assert len(q) == 3
        # tuple of expected test_id, expected_result values
        compare_set = set([(0, 20), (1, 20), (2, 75)])
        for t in q:
            val = t.test_id, t.expected_result
            assert val in compare_set
            compare_set.remove(val)
        assert len(compare_set) == 0

    @execute_count(2)
    def test_multiple_iterations_work_properly(self):
        """ Tests that iterating over a query set more than once works """
        # test with both the filtering method and the query method
        for q in (TestModel.objects(test_id=0), TestModel.objects(TestModel.test_id == 0)):
            # tuple of expected attempt_id, expected_result values
            compare_set = set([(0, 5), (1, 10), (2, 15), (3, 20)])
            for t in q:
                val = t.attempt_id, t.expected_result
                assert val in compare_set
                compare_set.remove(val)
            assert len(compare_set) == 0

            # try it again
            compare_set = set([(0, 5), (1, 10), (2, 15), (3, 20)])
            for t in q:
                val = t.attempt_id, t.expected_result
                assert val in compare_set
                compare_set.remove(val)
            assert len(compare_set) == 0

    @execute_count(2)
    def test_multiple_iterators_are_isolated(self):
        """
        tests that the use of one iterator does not affect the behavior of another
        """
        for q in (TestModel.objects(test_id=0), TestModel.objects(TestModel.test_id == 0)):
            q = q.order_by('attempt_id')
            expected_order = [0, 1, 2, 3]
            iter1 = iter(q)
            iter2 = iter(q)
            for attempt_id in expected_order:
                assert next(iter1).attempt_id == attempt_id
                assert next(iter2).attempt_id == attempt_id

    @execute_count(3)
    def test_get_success_case(self):
        """
        Tests that the .get() method works on new and existing querysets
        """
        m = TestModel.objects.get(test_id=0, attempt_id=0)
        assert isinstance(m, TestModel)
        assert m.test_id == 0
        assert m.attempt_id == 0

        q = TestModel.objects(test_id=0, attempt_id=0)
        m = q.get()
        assert isinstance(m, TestModel)
        assert m.test_id == 0
        assert m.attempt_id == 0

        q = TestModel.objects(test_id=0)
        m = q.get(attempt_id=0)
        assert isinstance(m, TestModel)
        assert m.test_id == 0
        assert m.attempt_id == 0

    @execute_count(3)
    def test_query_expression_get_success_case(self):
        """
        Tests that the .get() method works on new and existing querysets
        """
        m = TestModel.get(TestModel.test_id == 0, TestModel.attempt_id == 0)
        assert isinstance(m, TestModel)
        assert m.test_id == 0
        assert m.attempt_id == 0

        q = TestModel.objects(TestModel.test_id == 0, TestModel.attempt_id == 0)
        m = q.get()
        assert isinstance(m, TestModel)
        assert m.test_id == 0
        assert m.attempt_id == 0

        q = TestModel.objects(TestModel.test_id == 0)
        m = q.get(TestModel.attempt_id == 0)
        assert isinstance(m, TestModel)
        assert m.test_id == 0
        assert m.attempt_id == 0

    @execute_count(1)
    def test_get_doesnotexist_exception(self):
        """
        Tests that get calls that don't return a result raises a DoesNotExist error
        """
        with self.assertRaises(TestModel.DoesNotExist):
            TestModel.objects.get(test_id=100)

    @execute_count(1)
    def test_get_multipleobjects_exception(self):
        """
        Tests that get calls that return multiple results raise a MultipleObjectsReturned error
        """
        with self.assertRaises(TestModel.MultipleObjectsReturned):
            TestModel.objects.get(test_id=1)

    def test_allow_filtering_flag(self):
        """
        """

@execute_count(4)
def test_non_quality_filtering():
    class NonEqualityFilteringModel(Model):

        example_id = columns.UUID(primary_key=True, default=uuid.uuid4)
        sequence_id = columns.Integer(primary_key=True)  # sequence_id is a clustering key
        example_type = columns.Integer(index=True)
        created_at = columns.DateTime()

    drop_table(NonEqualityFilteringModel)
    sync_table(NonEqualityFilteringModel)

    # setup table, etc.

    NonEqualityFilteringModel.create(sequence_id=1, example_type=0, created_at=datetime.now())
    NonEqualityFilteringModel.create(sequence_id=3, example_type=0, created_at=datetime.now())
    NonEqualityFilteringModel.create(sequence_id=5, example_type=1, created_at=datetime.now())

    qa = NonEqualityFilteringModel.objects(NonEqualityFilteringModel.sequence_id > 3).allow_filtering()
    num = qa.count()
    assert num == 1, num


class TestQuerySetDistinct(BaseQuerySetUsage):

    @execute_count(1)
    def test_distinct_without_parameter(self):
        q = TestModel.objects.distinct()
        self.assertEqual(len(q), 3)

    @execute_count(1)
    def test_distinct_with_parameter(self):
        q = TestModel.objects.distinct(['test_id'])
        self.assertEqual(len(q), 3)

    @execute_count(1)
    def test_distinct_with_filter(self):
        q = TestModel.objects.distinct(['test_id']).filter(test_id__in=[1, 2])
        self.assertEqual(len(q), 2)

    @execute_count(1)
    def test_distinct_with_non_partition(self):
        with self.assertRaises(InvalidRequest):
            q = TestModel.objects.distinct(['description']).filter(test_id__in=[1, 2])
            len(q)

    @execute_count(1)
    def test_zero_result(self):
        q = TestModel.objects.distinct(['test_id']).filter(test_id__in=[52])
        self.assertEqual(len(q), 0)

    @greaterthancass21
    @execute_count(2)
    def test_distinct_with_explicit_count(self):
        q = TestModel.objects.distinct(['test_id'])
        self.assertEqual(q.count(), 3)

        q = TestModel.objects.distinct(['test_id']).filter(test_id__in=[1, 2])
        self.assertEqual(q.count(), 2)


class TestQuerySetOrdering(BaseQuerySetUsage):
    @execute_count(2)
    def test_order_by_success_case(self):
        q = TestModel.objects(test_id=0).order_by('attempt_id')
        expected_order = [0, 1, 2, 3]
        for model, expect in zip(q, expected_order):
            assert model.attempt_id == expect

        q = q.order_by('-attempt_id')
        expected_order.reverse()
        for model, expect in zip(q, expected_order):
            assert model.attempt_id == expect

    def test_ordering_by_non_second_primary_keys_fail(self):
        # kwarg filtering
        with self.assertRaises(query.QueryException):
            TestModel.objects(test_id=0).order_by('test_id')

        # kwarg filtering
        with self.assertRaises(query.QueryException):
            TestModel.objects(TestModel.test_id == 0).order_by('test_id')

    def test_ordering_by_non_primary_keys_fails(self):
        with self.assertRaises(query.QueryException):
            TestModel.objects(test_id=0).order_by('description')

    def test_ordering_on_indexed_columns_fails(self):
        with self.assertRaises(query.QueryException):
            IndexedTestModel.objects(test_id=0).order_by('attempt_id')

    @execute_count(8)
    def test_ordering_on_multiple_clustering_columns(self):
        TestMultiClusteringModel.create(one=1, two=1, three=4)
        TestMultiClusteringModel.create(one=1, two=1, three=2)
        TestMultiClusteringModel.create(one=1, two=1, three=5)
        TestMultiClusteringModel.create(one=1, two=1, three=1)
        TestMultiClusteringModel.create(one=1, two=1, three=3)

        results = TestMultiClusteringModel.objects.filter(one=1, two=1).order_by('-two', '-three')
        assert [r.three for r in results] == [5, 4, 3, 2, 1]

        results = TestMultiClusteringModel.objects.filter(one=1, two=1).order_by('two', 'three')
        assert [r.three for r in results] == [1, 2, 3, 4, 5]

        results = TestMultiClusteringModel.objects.filter(one=1, two=1).order_by('two').order_by('three')
        assert [r.three for r in results] == [1, 2, 3, 4, 5]


class TestQuerySetSlicing(BaseQuerySetUsage):

    @execute_count(1)
    def test_out_of_range_index_raises_error(self):
        q = TestModel.objects(test_id=0).order_by('attempt_id')
        with self.assertRaises(IndexError):
            q[10]

    @execute_count(1)
    def test_array_indexing_works_properly(self):
        q = TestModel.objects(test_id=0).order_by('attempt_id')
        expected_order = [0, 1, 2, 3]
        for i in range(len(q)):
            assert q[i].attempt_id == expected_order[i]

    @execute_count(1)
    def test_negative_indexing_works_properly(self):
        q = TestModel.objects(test_id=0).order_by('attempt_id')
        expected_order = [0, 1, 2, 3]
        assert q[-1].attempt_id == expected_order[-1]
        assert q[-2].attempt_id == expected_order[-2]

    @execute_count(1)
    def test_slicing_works_properly(self):
        q = TestModel.objects(test_id=0).order_by('attempt_id')
        expected_order = [0, 1, 2, 3]

        for model, expect in zip(q[1:3], expected_order[1:3]):
            self.assertEqual(model.attempt_id, expect)

        for model, expect in zip(q[0:3:2], expected_order[0:3:2]):
            self.assertEqual(model.attempt_id, expect)

    @execute_count(1)
    def test_negative_slicing(self):
        q = TestModel.objects(test_id=0).order_by('attempt_id')
        expected_order = [0, 1, 2, 3]

        for model, expect in zip(q[-3:], expected_order[-3:]):
            self.assertEqual(model.attempt_id, expect)

        for model, expect in zip(q[:-1], expected_order[:-1]):
            self.assertEqual(model.attempt_id, expect)

        for model, expect in zip(q[1:-1], expected_order[1:-1]):
            self.assertEqual(model.attempt_id, expect)

        for model, expect in zip(q[-3:-1], expected_order[-3:-1]):
            self.assertEqual(model.attempt_id, expect)

        for model, expect in zip(q[-3:-1:2], expected_order[-3:-1:2]):
            self.assertEqual(model.attempt_id, expect)


class TestQuerySetValidation(BaseQuerySetUsage):

    def test_primary_key_or_index_must_be_specified(self):
        """
        Tests that queries that don't have an equals relation to a primary key or indexed field fail
        """
        with self.assertRaises(query.QueryException):
            q = TestModel.objects(test_result=25)
            list([i for i in q])

    def test_primary_key_or_index_must_have_equal_relation_filter(self):
        """
        Tests that queries that don't have non equal (>,<, etc) relation to a primary key or indexed field fail
        """
        with self.assertRaises(query.QueryException):
            q = TestModel.objects(test_id__gt=0)
            list([i for i in q])

    @greaterthancass20
    @execute_count(7)
    def test_indexed_field_can_be_queried(self):
        """
        Tests that queries on an indexed field will work without any primary key relations specified
        """
        q = IndexedTestModel.objects(test_result=25)
        self.assertEqual(q.count(), 4)

        q = IndexedCollectionsTestModel.objects.filter(test_list__contains=42)
        self.assertEqual(q.count(), 1)

        q = IndexedCollectionsTestModel.objects.filter(test_list__contains=13)
        self.assertEqual(q.count(), 0)

        q = IndexedCollectionsTestModel.objects.filter(test_set__contains=42)
        self.assertEqual(q.count(), 1)

        q = IndexedCollectionsTestModel.objects.filter(test_set__contains=13)
        self.assertEqual(q.count(), 0)

        q = IndexedCollectionsTestModel.objects.filter(test_map__contains=42)
        self.assertEqual(q.count(), 1)

        q = IndexedCollectionsTestModel.objects.filter(test_map__contains=13)
        self.assertEqual(q.count(), 0)

    def test_custom_indexed_field_can_be_queried(self):
        """
        Tests that queries on an custom indexed field will work without any primary key relations specified
        """

        with self.assertRaises(query.QueryException):
            list(CustomIndexedTestModel.objects.filter(data='test'))  # not custom indexed

        # It should return InvalidRequest if target an indexed columns
        with self.assertRaises(InvalidRequest):
            list(CustomIndexedTestModel.objects.filter(indexed='test', data='test'))

        # It should return InvalidRequest if target an indexed columns
        with self.assertRaises(InvalidRequest):
            list(CustomIndexedTestModel.objects.filter(description='test', data='test'))

        # equals operator, server error since there is no real index, but it passes
        with self.assertRaises(InvalidRequest):
            list(CustomIndexedTestModel.objects.filter(description='test'))

        with self.assertRaises(InvalidRequest):
            list(CustomIndexedTestModel.objects.filter(test_id=1, description='test'))

        # gte operator, server error since there is no real index, but it passes
        # this can't work with a secondary index
        with self.assertRaises(InvalidRequest):
            list(CustomIndexedTestModel.objects.filter(description__gte='test'))

        with TestCluster().connect() as session:
            session.execute("CREATE INDEX custom_index_cqlengine ON {}.{} (description)".
                            format(DEFAULT_KEYSPACE, CustomIndexedTestModel._table_name))

        list(CustomIndexedTestModel.objects.filter(description='test'))
        list(CustomIndexedTestModel.objects.filter(test_id=1, description='test'))


class TestQuerySetDelete(BaseQuerySetUsage):

    @execute_count(9)
    def test_delete(self):
        TestModel.objects.create(test_id=3, attempt_id=0, description='try9', expected_result=50, test_result=40)
        TestModel.objects.create(test_id=3, attempt_id=1, description='try10', expected_result=60, test_result=40)
        TestModel.objects.create(test_id=3, attempt_id=2, description='try11', expected_result=70, test_result=45)
        TestModel.objects.create(test_id=3, attempt_id=3, description='try12', expected_result=75, test_result=45)

        assert TestModel.objects.count() == 16
        assert TestModel.objects(test_id=3).count() == 4

        TestModel.objects(test_id=3).delete()

        assert TestModel.objects.count() == 12
        assert TestModel.objects(test_id=3).count() == 0

    def test_delete_without_partition_key(self):
        """ Tests that attempting to delete a model without defining a partition key fails """
        with self.assertRaises(query.QueryException):
            TestModel.objects(attempt_id=0).delete()

    def test_delete_without_any_where_args(self):
        """ Tests that attempting to delete a whole table without any arguments will fail """
        with self.assertRaises(query.QueryException):
            TestModel.objects(attempt_id=0).delete()

    @greaterthanorequalcass30
    @execute_count(18)
    def test_range_deletion(self):
        """
        Tests that range deletion work as expected
        """

        for i in range(10):
            TestMultiClusteringModel.objects().create(one=1, two=i, three=i)

        TestMultiClusteringModel.objects(one=1, two__gte=0, two__lte=3).delete()
        self.assertEqual(6, len(TestMultiClusteringModel.objects.all()))

        TestMultiClusteringModel.objects(one=1, two__gt=3, two__lt=5).delete()
        self.assertEqual(5, len(TestMultiClusteringModel.objects.all()))

        TestMultiClusteringModel.objects(one=1, two__in=[8, 9]).delete()
        self.assertEqual(3, len(TestMultiClusteringModel.objects.all()))

        TestMultiClusteringModel.objects(one__in=[1], two__gte=0).delete()
        self.assertEqual(0, len(TestMultiClusteringModel.objects.all()))


class TimeUUIDQueryModel(Model):

    partition = columns.UUID(primary_key=True)
    time = columns.TimeUUID(primary_key=True)
    data = columns.Text(required=False)


class TestMinMaxTimeUUIDFunctions(BaseCassEngTestCase):
    @classmethod
    def setUpClass(cls):
        super(TestMinMaxTimeUUIDFunctions, cls).setUpClass()
        sync_table(TimeUUIDQueryModel)

    @classmethod
    def tearDownClass(cls):
        super(TestMinMaxTimeUUIDFunctions, cls).tearDownClass()
        drop_table(TimeUUIDQueryModel)

    @execute_count(7)
    def test_tzaware_datetime_support(self):
        """Test that using timezone aware datetime instances works with the
        MinTimeUUID/MaxTimeUUID functions.
        """
        pk = uuid4()
        midpoint_utc = datetime.utcnow().replace(tzinfo=TzOffset(0))
        midpoint_helsinki = midpoint_utc.astimezone(TzOffset(3))

        # Assert pre-condition that we have the same logical point in time
        assert midpoint_utc.utctimetuple() == midpoint_helsinki.utctimetuple()
        assert midpoint_utc.timetuple() != midpoint_helsinki.timetuple()

        TimeUUIDQueryModel.create(
            partition=pk,
            time=uuid_from_time(midpoint_utc - timedelta(minutes=1)),
            data='1')

        TimeUUIDQueryModel.create(
            partition=pk,
            time=uuid_from_time(midpoint_utc),
            data='2')

        TimeUUIDQueryModel.create(
            partition=pk,
            time=uuid_from_time(midpoint_utc + timedelta(minutes=1)),
            data='3')

        assert ['1', '2'] == [o.data for o in TimeUUIDQueryModel.filter(
            TimeUUIDQueryModel.partition == pk,
            TimeUUIDQueryModel.time <= functions.MaxTimeUUID(midpoint_utc))]

        assert ['1', '2'] == [o.data for o in TimeUUIDQueryModel.filter(
            TimeUUIDQueryModel.partition == pk,
            TimeUUIDQueryModel.time <= functions.MaxTimeUUID(midpoint_helsinki))]

        assert ['2', '3'] == [o.data for o in TimeUUIDQueryModel.filter(
            TimeUUIDQueryModel.partition == pk,
            TimeUUIDQueryModel.time >= functions.MinTimeUUID(midpoint_utc))]

        assert ['2', '3'] == [o.data for o in TimeUUIDQueryModel.filter(
            TimeUUIDQueryModel.partition == pk,
            TimeUUIDQueryModel.time >= functions.MinTimeUUID(midpoint_helsinki))]

    @execute_count(8)
    def test_success_case(self):
        """ Test that the min and max time uuid functions work as expected """
        pk = uuid4()
        startpoint = datetime.utcnow()
        TimeUUIDQueryModel.create(partition=pk, time=uuid_from_time(startpoint + timedelta(seconds=1)), data='1')
        TimeUUIDQueryModel.create(partition=pk, time=uuid_from_time(startpoint + timedelta(seconds=2)), data='2')
        midpoint = startpoint + timedelta(seconds=3)
        TimeUUIDQueryModel.create(partition=pk, time=uuid_from_time(startpoint + timedelta(seconds=4)), data='3')
        TimeUUIDQueryModel.create(partition=pk, time=uuid_from_time(startpoint + timedelta(seconds=5)), data='4')

        # test kwarg filtering
        q = TimeUUIDQueryModel.filter(partition=pk, time__lte=functions.MaxTimeUUID(midpoint))
        q = [d for d in q]
        self.assertEqual(len(q), 2, msg="Got: %s" % q)
        datas = [d.data for d in q]
        assert '1' in datas
        assert '2' in datas

        q = TimeUUIDQueryModel.filter(partition=pk, time__gte=functions.MinTimeUUID(midpoint))
        assert len(q) == 2
        datas = [d.data for d in q]
        assert '3' in datas
        assert '4' in datas

        # test query expression filtering
        q = TimeUUIDQueryModel.filter(
            TimeUUIDQueryModel.partition == pk,
            TimeUUIDQueryModel.time <= functions.MaxTimeUUID(midpoint)
        )
        q = [d for d in q]
        assert len(q) == 2
        datas = [d.data for d in q]
        assert '1' in datas
        assert '2' in datas

        q = TimeUUIDQueryModel.filter(
            TimeUUIDQueryModel.partition == pk,
            TimeUUIDQueryModel.time >= functions.MinTimeUUID(midpoint)
        )
        assert len(q) == 2
        datas = [d.data for d in q]
        assert '3' in datas
        assert '4' in datas


class TestInOperator(BaseQuerySetUsage):
    @execute_count(1)
    def test_kwarg_success_case(self):
        """ Tests the in operator works with the kwarg query method """
        q = TestModel.filter(test_id__in=[0, 1])
        assert q.count() == 8

    @execute_count(1)
    def test_query_expression_success_case(self):
        """ Tests the in operator works with the query expression query method """
        q = TestModel.filter(TestModel.test_id.in_([0, 1]))
        assert q.count() == 8

    @execute_count(5)
    def test_bool(self):
        """
        Adding coverage to cqlengine for bool types.

        @since 3.6
        @jira_ticket PYTHON-596
        @expected_result bool results should be filtered appropriately

        @test_category object_mapper
        """
        class bool_model(Model):
            k = columns.Integer(primary_key=True)
            b = columns.Boolean(primary_key=True)
            v = columns.Integer(default=3)
        sync_table(bool_model)

        bool_model.create(k=0, b=True)
        bool_model.create(k=0, b=False)
        self.assertEqual(len(bool_model.objects.all()), 2)
        self.assertEqual(len(bool_model.objects.filter(k=0, b=True)), 1)
        self.assertEqual(len(bool_model.objects.filter(k=0, b=False)), 1)

    @execute_count(3)
    def test_bool_filter(self):
        """
        Test to ensure that we don't translate boolean objects to String unnecessarily in filter clauses

        @since 3.6
        @jira_ticket PYTHON-596
        @expected_result We should not receive a server error

        @test_category object_mapper
        """
        class bool_model2(Model):
            k = columns.Boolean(primary_key=True)
            b = columns.Integer(primary_key=True)
            v = columns.Text()
        drop_table(bool_model2)
        sync_table(bool_model2)

        bool_model2.create(k=True, b=1, v='a')
        bool_model2.create(k=False, b=1, v='b')
        self.assertEqual(len(list(bool_model2.objects(k__in=(True, False)))), 2)


@greaterthancass20
class TestContainsOperator(BaseQuerySetUsage):

    @execute_count(6)
    def test_kwarg_success_case(self):
        """ Tests the CONTAINS operator works with the kwarg query method """
        q = IndexedCollectionsTestModel.filter(test_list__contains=1)
        self.assertEqual(q.count(), 2)

        q = IndexedCollectionsTestModel.filter(test_list__contains=13)
        self.assertEqual(q.count(), 0)

        q = IndexedCollectionsTestModel.filter(test_set__contains=3)
        self.assertEqual(q.count(), 2)

        q = IndexedCollectionsTestModel.filter(test_set__contains=13)
        self.assertEqual(q.count(), 0)

        q = IndexedCollectionsTestModel.filter(test_map__contains=42)
        self.assertEqual(q.count(), 1)

        q = IndexedCollectionsTestModel.filter(test_map__contains=13)
        self.assertEqual(q.count(), 0)

        with self.assertRaises(QueryException):
            q = IndexedCollectionsTestModel.filter(test_list_no_index__contains=1)
            self.assertEqual(q.count(), 0)
        with self.assertRaises(QueryException):
            q = IndexedCollectionsTestModel.filter(test_set_no_index__contains=1)
            self.assertEqual(q.count(), 0)
        with self.assertRaises(QueryException):
            q = IndexedCollectionsTestModel.filter(test_map_no_index__contains=1)
            self.assertEqual(q.count(), 0)

    @execute_count(6)
    def test_query_expression_success_case(self):
        """ Tests the CONTAINS operator works with the query expression query method """
        q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_list.contains_(1))
        self.assertEqual(q.count(), 2)

        q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_list.contains_(13))
        self.assertEqual(q.count(), 0)

        q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_set.contains_(3))
        self.assertEqual(q.count(), 2)

        q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_set.contains_(13))
        self.assertEqual(q.count(), 0)

        q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_map.contains_(42))
        self.assertEqual(q.count(), 1)

        q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_map.contains_(13))
        self.assertEqual(q.count(), 0)

        with self.assertRaises(QueryException):
            q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_map_no_index.contains_(1))
            self.assertEqual(q.count(), 0)
        with self.assertRaises(QueryException):
            q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_map_no_index.contains_(1))
            self.assertEqual(q.count(), 0)
        with self.assertRaises(QueryException):
            q = IndexedCollectionsTestModel.filter(IndexedCollectionsTestModel.test_map_no_index.contains_(1))
            self.assertEqual(q.count(), 0)


class TestValuesList(BaseQuerySetUsage):

    @execute_count(2)
    def test_values_list(self):
        q = TestModel.objects.filter(test_id=0, attempt_id=1)
        item = q.values_list('test_id', 'attempt_id', 'description', 'expected_result', 'test_result').first()
        assert item == [0, 1, 'try2', 10, 30]

        item = q.values_list('expected_result', flat=True).first()
        assert item == 10


class TestObjectsProperty(BaseQuerySetUsage):
    @execute_count(1)
    def test_objects_property_returns_fresh_queryset(self):
        assert TestModel.objects._result_cache is None
        len(TestModel.objects)  # evaluate queryset
        assert TestModel.objects._result_cache is None


class PageQueryTests(BaseCassEngTestCase):
    @execute_count(3)
    def test_paged_result_handling(self):
        if PROTOCOL_VERSION < 2:
            raise unittest.SkipTest("Paging requires native protocol 2+, currently using: {0}".format(PROTOCOL_VERSION))

        # addresses #225
        class PagingTest(Model):
            id = columns.Integer(primary_key=True)
            val = columns.Integer()
        sync_table(PagingTest)

        PagingTest.create(id=1, val=1)
        PagingTest.create(id=2, val=2)

        session = get_session()
        with mock.patch.object(session, 'default_fetch_size', 1):
            results = PagingTest.objects()[:]

        assert len(results) == 2


class ModelQuerySetTimeoutTestCase(BaseQuerySetUsage):
    def test_default_timeout(self):
        with mock.patch.object(Session, 'execute') as mock_execute:
            list(TestModel.objects())
            self.assertEqual(mock_execute.call_args[-1]['timeout'], NOT_SET)

    def test_float_timeout(self):
        with mock.patch.object(Session, 'execute') as mock_execute:
            list(TestModel.objects().timeout(0.5))
            self.assertEqual(mock_execute.call_args[-1]['timeout'], 0.5)

    def test_none_timeout(self):
        with mock.patch.object(Session, 'execute') as mock_execute:
            list(TestModel.objects().timeout(None))
            self.assertEqual(mock_execute.call_args[-1]['timeout'], None)


class DMLQueryTimeoutTestCase(BaseQuerySetUsage):
    def setUp(self):
        self.model = TestModel(test_id=1, attempt_id=1, description='timeout test')
        super(DMLQueryTimeoutTestCase, self).setUp()

    def test_default_timeout(self):
        with mock.patch.object(Session, 'execute') as mock_execute:
            self.model.save()
            self.assertEqual(mock_execute.call_args[-1]['timeout'], NOT_SET)

    def test_float_timeout(self):
        with mock.patch.object(Session, 'execute') as mock_execute:
            self.model.timeout(0.5).save()
            self.assertEqual(mock_execute.call_args[-1]['timeout'], 0.5)

    def test_none_timeout(self):
        with mock.patch.object(Session, 'execute') as mock_execute:
            self.model.timeout(None).save()
            self.assertEqual(mock_execute.call_args[-1]['timeout'], None)

    def test_timeout_then_batch(self):
        b = query.BatchQuery()
        m = self.model.timeout(None)
        with self.assertRaises(AssertionError):
            m.batch(b)

    def test_batch_then_timeout(self):
        b = query.BatchQuery()
        m = self.model.batch(b)
        with self.assertRaises(AssertionError):
            m.timeout(0.5)


class DBFieldModel(Model):
    k0 = columns.Integer(partition_key=True, db_field='a')
    k1 = columns.Integer(partition_key=True, db_field='b')
    c0 = columns.Integer(primary_key=True, db_field='c')
    v0 = columns.Integer(db_field='d')
    v1 = columns.Integer(db_field='e', index=True)


class DBFieldModelMixed1(Model):
    k0 = columns.Integer(partition_key=True, db_field='a')
    k1 = columns.Integer(partition_key=True)
    c0 = columns.Integer(primary_key=True, db_field='c')
    v0 = columns.Integer(db_field='d')
    v1 = columns.Integer(index=True)


class DBFieldModelMixed2(Model):
    k0 = columns.Integer(partition_key=True)
    k1 = columns.Integer(partition_key=True, db_field='b')
    c0 = columns.Integer(primary_key=True)
    v0 = columns.Integer(db_field='d')
    v1 = columns.Integer(index=True, db_field='e')


class TestModelQueryWithDBField(BaseCassEngTestCase):

    def setUp(cls):
        super(TestModelQueryWithDBField, cls).setUpClass()
        cls.model_list = [DBFieldModel, DBFieldModelMixed1, DBFieldModelMixed2]
        for model in cls.model_list:
            sync_table(model)

    def tearDown(cls):
        super(TestModelQueryWithDBField, cls).tearDownClass()
        for model in cls.model_list:
            drop_table(model)

    @execute_count(33)
    def test_basic_crud(self):
        """
        Tests creation update and delete of object model queries that are using db_field mappings.

        @since 3.1
        @jira_ticket PYTHON-351
        @expected_result results are properly retrieved without errors

        @test_category object_mapper
        """
        for model in self.model_list:
            values = {'k0': 1, 'k1': 2, 'c0': 3, 'v0': 4, 'v1': 5}
            # create
            i = model.create(**values)
            i = model.objects(k0=i.k0, k1=i.k1).first()
            self.assertEqual(i, model(**values))

            # create
            values['v0'] = 101
            i.update(v0=values['v0'])
            i = model.objects(k0=i.k0, k1=i.k1).first()
            self.assertEqual(i, model(**values))

            # delete
            model.objects(k0=i.k0, k1=i.k1).delete()
            i = model.objects(k0=i.k0, k1=i.k1).first()
            self.assertIsNone(i)

            i = model.create(**values)
            i = model.objects(k0=i.k0, k1=i.k1).first()
            self.assertEqual(i, model(**values))
            i.delete()
            model.objects(k0=i.k0, k1=i.k1).delete()
            i = model.objects(k0=i.k0, k1=i.k1).first()
            self.assertIsNone(i)

    @execute_count(21)
    def test_slice(self):
        """
        Tests slice queries for object models that are using db_field mapping

        @since 3.1
        @jira_ticket PYTHON-351
        @expected_result results are properly retrieved without errors

        @test_category object_mapper
        """
        for model in self.model_list:
            values = {'k0': 1, 'k1': 3, 'c0': 3, 'v0': 4, 'v1': 5}
            clustering_values = range(3)
            for c in clustering_values:
                values['c0'] = c
                i = model.create(**values)

            self.assertEqual(model.objects(k0=i.k0, k1=i.k1).count(), len(clustering_values))
            self.assertEqual(model.objects(k0=i.k0, k1=i.k1, c0=i.c0).count(), 1)
            self.assertEqual(model.objects(k0=i.k0, k1=i.k1, c0__lt=i.c0).count(), len(clustering_values[:-1]))
            self.assertEqual(model.objects(k0=i.k0, k1=i.k1, c0__gt=0).count(), len(clustering_values[1:]))

    @execute_count(15)
    def test_order(self):
        """
        Tests order by queries for object models that are using db_field mapping

        @since 3.1
        @jira_ticket PYTHON-351
        @expected_result results are properly retrieved without errors

        @test_category object_mapper
        """
        for model in self.model_list:
            values = {'k0': 1, 'k1': 4, 'c0': 3, 'v0': 4, 'v1': 5}
            clustering_values = range(3)
            for c in clustering_values:
                values['c0'] = c
                i = model.create(**values)
            self.assertEqual(model.objects(k0=i.k0, k1=i.k1).order_by('c0').first().c0, clustering_values[0])
            self.assertEqual(model.objects(k0=i.k0, k1=i.k1).order_by('-c0').first().c0, clustering_values[-1])

    @execute_count(15)
    def test_index(self):
        """
        Tests queries using index fields for object models using db_field mapping

        @since 3.1
        @jira_ticket PYTHON-351
        @expected_result results are properly retrieved without errors

        @test_category object_mapper
        """
        for model in self.model_list:
            values = {'k0': 1, 'k1': 5, 'c0': 3, 'v0': 4, 'v1': 5}
            clustering_values = range(3)
            for c in clustering_values:
                values['c0'] = c
                values['v1'] = c
                i = model.create(**values)
            self.assertEqual(model.objects(k0=i.k0, k1=i.k1).count(), len(clustering_values))
            self.assertEqual(model.objects(k0=i.k0, k1=i.k1, v1=0).count(), 1)

    @execute_count(1)
    def test_db_field_names_used(self):
        """
        Tests to ensure that with generated cql update statements correctly utilize the db_field values.

        @since 3.2
        @jira_ticket PYTHON-530
        @expected_result resulting cql_statements will use the db_field values

        @test_category object_mapper
        """

        values = ('k0', 'k1', 'c0', 'v0', 'v1')
        # Test QuerySet Path
        b = BatchQuery()
        DBFieldModel.objects(k0=1).batch(b).update(
            v0=0,
            v1=9,
        )
        for value in values:
            self.assertTrue(value not in str(b.queries[0]))

        # Test DML path
        b2 = BatchQuery()
        dml_field_model = DBFieldModel.create(k0=1, k1=5, c0=3, v0=4, v1=5)
        dml_field_model.batch(b2).update(
            v0=0,
            v1=9,
        )
        for value in values:
            self.assertTrue(value not in str(b2.queries[0]))

    def test_db_field_value_list(self):
        DBFieldModel.create(k0=0, k1=0, c0=0, v0=4, v1=5)

        self.assertEqual(DBFieldModel.objects.filter(c0=0, k0=0, k1=0).values_list('c0', 'v0')._defer_fields,
                         {'a', 'c', 'b'})
        self.assertEqual(DBFieldModel.objects.filter(c0=0, k0=0, k1=0).values_list('c0', 'v0')._only_fields,
                         ['c', 'd'])

        list(DBFieldModel.objects.filter(c0=0, k0=0, k1=0).values_list('c0', 'v0'))

class TestModelSmall(Model):

    test_id = columns.Integer(primary_key=True)


class TestModelQueryWithFetchSize(BaseCassEngTestCase):
    """
    Test FetchSize, and ensure that results are returned correctly
    regardless of the paging size

    @since 3.1
    @jira_ticket PYTHON-324
    @expected_result results are properly retrieved and the correct size

    @test_category object_mapper
    """

    @classmethod
    def setUpClass(cls):
        super(TestModelQueryWithFetchSize, cls).setUpClass()
        sync_table(TestModelSmall)

    @classmethod
    def tearDownClass(cls):
        super(TestModelQueryWithFetchSize, cls).tearDownClass()
        drop_table(TestModelSmall)

    @execute_count(19)
    def test_defaultFetchSize(self):
        # Use smaller batch sizes to avoid hitting the max.  We trigger an InvalidRequest
        # response for Cassandra 4.1.x and 5.0.x if we just do the whole thing as one
        # large batch.  We're just using this to populate values for a test, however,
        # so shifting to smaller batches should be fine.
        for i in range(0, 5000, 500):
            with BatchQuery() as b:
                range_max = i + 500
                for j in range(i, range_max):
                    TestModelSmall.batch(b).create(test_id=j)
        with BatchQuery() as b:
            for i in range(5000, 5100):
                TestModelSmall.batch(b).create(test_id=i)

        self.assertEqual(len(TestModelSmall.objects.fetch_size(1)), 5100)
        self.assertEqual(len(TestModelSmall.objects.fetch_size(500)), 5100)
        self.assertEqual(len(TestModelSmall.objects.fetch_size(4999)), 5100)
        self.assertEqual(len(TestModelSmall.objects.fetch_size(5000)), 5100)
        self.assertEqual(len(TestModelSmall.objects.fetch_size(5001)), 5100)
        self.assertEqual(len(TestModelSmall.objects.fetch_size(5100)), 5100)
        self.assertEqual(len(TestModelSmall.objects.fetch_size(5101)), 5100)
        self.assertEqual(len(TestModelSmall.objects.fetch_size(1)), 5100)

        with self.assertRaises(QueryException):
            TestModelSmall.objects.fetch_size(0)
        with self.assertRaises(QueryException):
            TestModelSmall.objects.fetch_size(-1)


class People(Model):
    __table_name__ = "people"
    last_name = columns.Text(primary_key=True, partition_key=True)
    first_name = columns.Text(primary_key=True)
    birthday = columns.DateTime()


class People2(Model):
    __table_name__ = "people"
    last_name = columns.Text(primary_key=True, partition_key=True)
    first_name = columns.Text(primary_key=True)
    middle_name = columns.Text()
    birthday = columns.DateTime()


class TestModelQueryWithDifferedFeld(BaseCassEngTestCase):
    """
    Tests that selects with filter will deffer population of known values until after the results are returned.
    I.E.  Instead of generating SELECT * FROM People WHERE last_name="Smith" It will generate
    SELECT first_name, birthday FROM People WHERE last_name="Smith"
    Where last_name 'smith' will populated post query

    @since 3.2
    @jira_ticket PYTHON-520
    @expected_result only needed fields are included in the query

    @test_category object_mapper
    """
    @classmethod
    def setUpClass(cls):
        super(TestModelQueryWithDifferedFeld, cls).setUpClass()
        sync_table(People)

    @classmethod
    def tearDownClass(cls):
        super(TestModelQueryWithDifferedFeld, cls).tearDownClass()
        drop_table(People)

    @execute_count(8)
    def test_defaultFetchSize(self):
        # Populate Table
        People.objects.create(last_name="Smith", first_name="John", birthday=datetime.now())
        People.objects.create(last_name="Bestwater", first_name="Alan", birthday=datetime.now())
        People.objects.create(last_name="Smith", first_name="Greg", birthday=datetime.now())
        People.objects.create(last_name="Smith", first_name="Adam", birthday=datetime.now())

        # Check query constructions
        expected_fields = ['first_name', 'birthday']
        self.assertEqual(People.filter(last_name="Smith")._select_fields(), expected_fields)
        # Validate correct fields are fetched
        smiths = list(People.filter(last_name="Smith"))
        self.assertEqual(len(smiths), 3)
        self.assertTrue(smiths[0].last_name is not None)

        # Modify table with new value
        sync_table(People2)

        # populate new format
        People2.objects.create(last_name="Smith", first_name="Chris", middle_name="Raymond", birthday=datetime.now())
        People2.objects.create(last_name="Smith", first_name="Andrew", middle_name="Micheal", birthday=datetime.now())

        # validate query construction
        expected_fields = ['first_name', 'middle_name', 'birthday']
        self.assertEqual(People2.filter(last_name="Smith")._select_fields(), expected_fields)

        # validate correct items are returneds
        smiths = list(People2.filter(last_name="Smith"))
        self.assertEqual(len(smiths), 5)
        self.assertTrue(smiths[0].last_name is not None)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tests.integration import USE_CASS_EXTERNAL, use_cluster, TestCluster


class MisconfiguredAuthenticationTests(unittest.TestCase):
    """ One node (not the contact point) has password auth. The rest of the nodes have no auth """
    @classmethod
    def setUpClass(cls):
        if not USE_CASS_EXTERNAL:
            ccm_cluster = use_cluster(cls.__name__, [3], start=False)
            node3 = ccm_cluster.nodes['node3']
            node3.set_configuration_options(values={
                'authenticator': 'PasswordAuthenticator',
                'authorizer': 'CassandraAuthorizer',
            })
            ccm_cluster.start(wait_for_binary_proto=True)

            cls.ccm_cluster = ccm_cluster

    def test_connect_no_auth_provider(self):
        cluster = TestCluster()
        cluster.connect()
        cluster.refresh_nodes()
        down_hosts = [host for host in cluster.metadata.all_hosts() if not host.is_up]
        self.assertEqual(len(down_hosts), 1)
        cluster.shutdown()

    @classmethod
    def tearDownClass(cls):
        if not USE_CASS_EXTERNAL:
            cls.ccm_cluster.stop()



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, socket, errno
from ccmlib import common

from cassandra.cluster import NoHostAvailable
from cassandra.io.asyncorereactor import AsyncoreConnection

from tests import is_monkey_patched
from tests.integration import use_cluster, remove_cluster, TestCluster

if is_monkey_patched():
    LibevConnection = -1
    AsyncoreConnection = -1
else:
    try:
        from cassandra.io.libevreactor import LibevConnection
    except ImportError:
        LibevConnection = None

import unittest


# If more modules do IPV6 testing, this can be moved down to integration.__init__.
# For now, just keeping the clutter here
IPV6_CLUSTER_NAME = 'ipv6_test_cluster'


def setup_module(module):
    if os.name != "nt":
        validate_host_viable()
        # We use a dedicated cluster (instead of common singledc, as in other tests) because
        # it's most likely that the test host will only have one local ipv6 address (::1)
        # singledc has three
        use_cluster(IPV6_CLUSTER_NAME, [1], ipformat='::%d')


def teardown_module():
    remove_cluster()


def validate_ccm_viable():
    try:
        common.normalize_interface(('::1', 0))
    except:
        raise unittest.SkipTest('this version of ccm does not support ipv6')


def validate_host_viable():
    # this is something ccm does when starting, but preemptively check to avoid
    # spinning up the cluster if it's not going to work
    try:
        common.assert_socket_available(('::1', 9042))
    except:
        raise unittest.SkipTest('failed binding ipv6 loopback ::1 on 9042')


class IPV6ConnectionTest(object):

    connection_class = None

    def test_connect(self):
        cluster = TestCluster(connection_class=self.connection_class, contact_points=['::1'], connect_timeout=10)
        session = cluster.connect()
        future = session.execute_async("SELECT * FROM system.local")
        future.result()
        self.assertEqual(future._current_host.address, '::1')
        cluster.shutdown()

    def test_error(self):
        cluster = TestCluster(connection_class=self.connection_class, contact_points=['::1'], port=9043,
                              connect_timeout=10)
        self.assertRaisesRegex(NoHostAvailable, '\(\'Unable to connect.*%s.*::1\', 9043.*Connection refused.*'
                                % errno.ECONNREFUSED, cluster.connect)

    def test_error_multiple(self):
        if len(socket.getaddrinfo('localhost', 9043, socket.AF_UNSPEC, socket.SOCK_STREAM)) < 2:
            raise unittest.SkipTest('localhost only resolves one address')
        cluster = TestCluster(connection_class=self.connection_class, contact_points=['localhost'], port=9043,
                              connect_timeout=10)
        self.assertRaisesRegex(NoHostAvailable, '\(\'Unable to connect.*Tried connecting to \[\(.*\(.*\].*Last error',
                                cluster.connect)


class LibevConnectionTests(IPV6ConnectionTest, unittest.TestCase):

    connection_class = LibevConnection

    def setUp(self):
        if os.name == "nt":
            raise unittest.SkipTest("IPv6 is currently not supported under Windows")

        if LibevConnection == -1:
            raise unittest.SkipTest("Can't test libev with monkey patching")
        elif LibevConnection is None:
            raise unittest.SkipTest("Libev does not appear to be installed properly")


class AsyncoreConnectionTests(IPV6ConnectionTest, unittest.TestCase):

    connection_class = AsyncoreConnection

    def setUp(self):
        if os.name == "nt":
            raise unittest.SkipTest("IPv6 is currently not supported under Windows")

        if AsyncoreConnection == -1:
            raise unittest.SkipTest("Can't test asyncore with monkey patching")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from packaging.version import Version

from tests import notwindows
from tests.unit.cython.utils import notcython
from tests.integration import (execute_until_pass,
                               execute_with_long_wait_retry, use_cluster, TestCluster)

import unittest


CCM_IS_DSE = (os.environ.get('CCM_IS_DSE', None) == 'true')


@unittest.skipIf(os.environ.get('CCM_ARGS', None), 'environment has custom CCM_ARGS; skipping')
@notwindows
@notcython  # no need to double up on this test; also __default__ setting doesn't work
class DseCCMClusterTest(unittest.TestCase):
    """
    This class can be executed setting the DSE_VERSION variable, for example:
    DSE_VERSION=5.1.4 python2.7 -m nose tests/integration/standard/test_dse.py
    If CASSANDRA_VERSION is set instead, it will be converted to the corresponding DSE_VERSION
    """

    def test_dse_5x(self):
        self._test_basic(Version('5.1.10'))

    def test_dse_60(self):
        self._test_basic(Version('6.0.2'))

    @unittest.skipUnless(CCM_IS_DSE, 'DSE version unavailable')
    def test_dse_67(self):
        self._test_basic(Version('6.7.0'))

    def _test_basic(self, dse_version):
        """
        Test basic connection and usage
        """
        cluster_name = '{}-{}'.format(
            self.__class__.__name__, dse_version.base_version.replace('.', '_')
        )
        use_cluster(cluster_name=cluster_name, nodes=[3], dse_options={})

        cluster = TestCluster()
        session = cluster.connect()
        result = execute_until_pass(
            session,
            """
            CREATE KEYSPACE clustertests
            WITH replication = {'class': 'SimpleStrategy', 'replication_factor': '1'}
            """)
        self.assertFalse(result)

        result = execute_with_long_wait_retry(
            session,
            """
            CREATE TABLE clustertests.cf0 (
                a text,
                b text,
                c text,
                PRIMARY KEY (a, b)
            )
            """)
        self.assertFalse(result)

        result = session.execute(
            """
            INSERT INTO clustertests.cf0 (a, b, c) VALUES ('a', 'b', 'c')
            """)
        self.assertFalse(result)

        result = session.execute("SELECT * FROM clustertests.cf0")
        self.assertEqual([('a', 'b', 'c')], result)

        execute_with_long_wait_retry(session, "DROP KEYSPACE clustertests")

        cluster.shutdown()



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import struct
import math
from cassandra.cqltypes import lookup_casstype
from cassandra.protocol import ProtocolVersion
from cassandra.cqltypes import PointType, LineStringType, PolygonType, WKBGeometryType
from cassandra.util import Point, LineString, Polygon, _LinearRing, Distance, _HAS_GEOMET

wkb_be = 0
wkb_le = 1

protocol_versions = ProtocolVersion.SUPPORTED_VERSIONS


class GeoTypes(unittest.TestCase):

    samples = (Point(1, 2), LineString(((1, 2), (3, 4), (5, 6))), Polygon([(10.1, 10.0), (110.0, 10.0), (110., 110.0), (10., 110.0), (10., 10.0)], [[(20., 20.0), (20., 30.0), (30., 30.0), (30., 20.0), (20., 20.0)], [(40., 20.0), (40., 30.0), (50., 30.0), (50., 20.0), (40., 20.0)]]))

    def test_marshal_platform(self):
        for proto_ver in protocol_versions:
            for geo in self.samples:
                cql_type = lookup_casstype(geo.__class__.__name__ + 'Type')
                self.assertEqual(cql_type.from_binary(cql_type.to_binary(geo, proto_ver), proto_ver), geo)

    def _verify_both_endian(self, typ, body_fmt, params, expected):
        for proto_ver in protocol_versions:
            self.assertEqual(typ.from_binary(struct.pack(">BI" + body_fmt, wkb_be, *params), proto_ver), expected)
            self.assertEqual(typ.from_binary(struct.pack("<BI" + body_fmt, wkb_le, *params), proto_ver), expected)

    def test_both_endian(self):
        self._verify_both_endian(PointType, "dd", (WKBGeometryType.POINT, 1, 2), Point(1, 2))
        self._verify_both_endian(LineStringType, "Idddddd", (WKBGeometryType.LINESTRING, 3, 1, 2, 3, 4, 5, 6), LineString(((1, 2), (3, 4), (5, 6))))
        self._verify_both_endian(PolygonType, "IIdddddd", (WKBGeometryType.POLYGON, 1, 3, 1, 2, 3, 4, 5, 6), Polygon(((1, 2), (3, 4), (5, 6))))

    def test_empty_wkb(self):
        for cls in (LineString, Polygon):
            class_name = cls.__name__
            cql_type = lookup_casstype(class_name + 'Type')
            self.assertEqual(str(cql_type.from_binary(cql_type.to_binary(cls(), 0), 0)), class_name.upper() + " EMPTY")
        self.assertEqual(str(PointType.from_binary(PointType.to_binary(Point(), 0), 0)), "POINT (nan nan)")

    def test_str_wkt(self):
        self.assertEqual(str(Point(1., 2.)), 'POINT (1.0 2.0)')
        self.assertEqual(str(Point()), "POINT (nan nan)")
        self.assertEqual(str(LineString(((1., 2.), (3., 4.), (5., 6.)))), 'LINESTRING (1.0 2.0, 3.0 4.0, 5.0 6.0)')
        self.assertEqual(str(_LinearRing(((1., 2.), (3., 4.), (5., 6.)))), 'LINEARRING (1.0 2.0, 3.0 4.0, 5.0 6.0)')
        self.assertEqual(str(Polygon([(10.1, 10.0), (110.0, 10.0), (110., 110.0), (10., 110.0), (10., 10.0)],
                                     [[(20., 20.0), (20., 30.0), (30., 30.0), (30., 20.0), (20., 20.0)],
                                      [(40., 20.0), (40., 30.0), (50., 30.0), (50., 20.0), (40., 20.0)]])),
                         'POLYGON ((10.1 10.0, 110.0 10.0, 110.0 110.0, 10.0 110.0, 10.0 10.0), (20.0 20.0, 20.0 30.0, 30.0 30.0, 30.0 20.0, 20.0 20.0), (40.0 20.0, 40.0 30.0, 50.0 30.0, 50.0 20.0, 40.0 20.0))')

        class LinearRing(_LinearRing):
            pass
        for cls in (LineString, LinearRing, Polygon):
            self.assertEqual(str(cls()), cls.__name__.upper() + " EMPTY")

    def test_repr(self):
        for geo in (Point(1., 2.),
                    LineString(((1., 2.), (3., 4.), (5., 6.))),
                    _LinearRing(((1., 2.), (3., 4.), (5., 6.))),
                    Polygon([(10.1, 10.0), (110.0, 10.0), (110., 110.0), (10., 110.0), (10., 10.0)],
                            [[(20., 20.0), (20., 30.0), (30., 30.0), (30., 20.0), (20., 20.0)],
                             [(40., 20.0), (40., 30.0), (50., 30.0), (50., 20.0), (40., 20.0)]])):
            self.assertEqual(eval(repr(geo)), geo)

    def test_hash(self):
        for geo in (Point(1., 2.),
                    LineString(((1., 2.), (3., 4.), (5., 6.))),
                    _LinearRing(((1., 2.), (3., 4.), (5., 6.))),
                    Polygon([(10.1, 10.0), (110.0, 10.0), (110., 110.0), (10., 110.0), (10., 10.0)],
                            [[(20., 20.0), (20., 30.0), (30., 30.0), (30., 20.0), (20., 20.0)],
                             [(40., 20.0), (40., 30.0), (50., 30.0), (50., 20.0), (40., 20.0)]])):
            self.assertEqual(len(set((geo, geo))), 1)

    def test_eq(self):
        for geo in (Point(1., 2.),
                    LineString(((1., 2.), (3., 4.), (5., 6.))),
                    _LinearRing(((1., 2.), (3., 4.), (5., 6.))),
                    Polygon([(10.1, 10.0), (110.0, 10.0), (110., 110.0), (10., 110.0), (10., 10.0)],
                            [[(20., 20.0), (20., 30.0), (30., 30.0), (30., 20.0), (20., 20.0)],
                             [(40., 20.0), (40., 30.0), (50., 30.0), (50., 20.0), (40., 20.0)]])):
            # same type
            self.assertEqual(geo, geo)

            # does not blow up on other types
            # specifically use assertFalse(eq) to make sure we're using the geo __eq__ operator
            self.assertFalse(geo == object())

@unittest.skipUnless(_HAS_GEOMET, "Skip wkt geometry tests when geomet is not installed")
class WKTTest(unittest.TestCase):

    def test_line_parse(self):
        """
        This test exercises the parsing logic our LINESTRING WKT object
        @since 1.2
        @jira_ticket PYTHON-641
        @test_category dse geometric
        @expected_result We should be able to form LINESTRINGS objects from properly formatted WKT strings
        """

        # Test simple line string
        ls = "LINESTRING (1.0 2.0, 3.0 4.0, 5.0 6.0)"
        lo = LineString.from_wkt(ls)
        lo_expected_cords = ((1.0, 2.0), (3.0, 4.0), (5.0, 6.0))
        self.assertEqual(lo.coords, lo_expected_cords)

        # Test very long line string
        long_ls = self._construct_line_string(10000)
        long_lo = LineString.from_wkt(long_ls)
        self.assertEqual(len(long_lo.coords), 10000)
        self.assertEqual(long_lo.coords, self._construct_line_string_expected_cords(10000))

        # Test line string with negative numbers
        ls = "LINESTRING (-1.3 1.2, 3.23 -4.54, 1.34 -9.26)"
        lo = LineString.from_wkt(ls)
        lo_expected_cords = ((-1.3, 1.2), (3.23, -4.54), (1.34, -9.26))
        self.assertEqual(lo.coords, lo_expected_cords)

        # Test bad line strings
        bls = "LINESTRIN (1.0 2.0, 3.0 4.0, 5.0 6.0)"
        with self.assertRaises(ValueError):
            blo = LineString.from_wkt(bls)
        bls = "LINESTRING (1.0 2.0 3.0 4.0 5.0"
        with self.assertRaises(ValueError):
            blo = LineString.from_wkt(bls)

        # Test with NAN
        ls = "LINESTRING (NAN NAN, NAN NAN)"
        lo = LineString.from_wkt(ls)
        self.assertEqual(len(lo.coords), 2)
        for cords in lo.coords:
            for cord in cords:
                self.assertTrue(math.isnan(cord))

    def test_distance_parse(self):
        """
        This test exercises the parsing logic our Distance WKT object
        @since 1.2
        @jira_ticket PYTHON-670
        @test_category dse geometric
        @expected_result We should be able to form Distance objects from properly formatted WKT strings
        """

        ds = "DISTANCE ((12, 10) 3)"
        do = Distance(12, 10, 3)
        self.assertEqual(do.x, 12)
        self.assertEqual(do.y, 10)
        self.assertEqual(do.radius, 3)
        # Test bad distance strings

        bds = "DISTANCE ((1.0 2.0))"
        with self.assertRaises(ValueError):
            bdo = Distance.from_wkt(bds)
        bps = "DISTANCE ((1.0 2.0 3.0 4.0 5.0)"
        with self.assertRaises(ValueError):
            bdo = Distance.from_wkt(bds)

        # NAN isn't supported, truncating not supported

    def test_point_parse(self):
        """
        This test exercises the parsing logic our POINT WKT object
        @since 1.2
        @jira_ticket PYTHON-641
        @test_category dse geometric
        @expected_result We should be able to form POINT objects from properly formatted WKT strings
        """

        # Test basic point
        ps = "POINT (1.0 2.0)"
        po = Point.from_wkt(ps)
        self.assertEqual(po.x, 1.0)
        self.assertEqual(po.y, 2.0)

        # Test bad point strings
        bps = "POIN (1.0 2.0)"
        with self.assertRaises(ValueError):
            bpo = Point.from_wkt(bps)
        bps = "POINT (1.0 2.0 3.0 4.0 5.0"
        with self.assertRaises(ValueError):
            bpo = Point.from_wkt(bps)

        # Points get truncated automatically
        tps = "POINT (9.0 2.0 3.0 4.0 5.0)"
        tpo = Point.from_wkt(tps)
        self.assertEqual(tpo.x, 9.0)
        self.assertEqual(tpo.y, 2.0)

        # Test point with NAN
        ps = "POINT (NAN NAN)"
        po = Point.from_wkt(ps)
        self.assertTrue(math.isnan(po.x))
        self.assertTrue(math.isnan(po.y))

    def test_polygon_parse(self):
        """
        This test exercises the parsing logic our POLYGON WKT object
        @since 1.2
        @jira_ticket PYTHON-641
        @test_category dse geometric
        @expected_result We should be able to form POLYGON objects from properly formatted WKT strings
        """

        example_poly_string = 'POLYGON ((10.1 10.0, 110.0 10.0, 110.0 110.0, 10.0 110.0, 10.0 10.0), (20.0 20.0, 20.0 30.0, 30.0 30.0, 30.0 20.0, 20.0 20.0), (40.0 20.0, 40.0 30.0, 50.0 30.0, 50.0 20.0, 40.0 20.0))'
        poly_obj = Polygon.from_wkt(example_poly_string)
        expected_ex_coords = ((10.1, 10.0), (110.0, 10.0), (110.0, 110.0), (10.0, 110.0), (10.0, 10.0))
        expected_in_coords_1 = ((20.0, 20.0), (20.0, 30.0), (30.0, 30.0), (30.0, 20.0), (20.0, 20.0))
        expected_in_coords_2 = ((40.0, 20.0), (40.0, 30.0), (50.0, 30.0), (50.0, 20.0), (40.0, 20.0))
        self.assertEqual(poly_obj.exterior.coords, expected_ex_coords)
        self.assertEqual(len(poly_obj.interiors), 2)
        self.assertEqual(poly_obj.interiors[0].coords, expected_in_coords_1)
        self.assertEqual(poly_obj.interiors[1].coords, expected_in_coords_2)

        # Test with very long polygon
        long_poly_string = self._construct_polygon_string(10000)
        long_poly_obj = Polygon.from_wkt(long_poly_string)
        self.assertEqual(len(long_poly_obj.exterior.coords), 10000)
        #for expected, recieved in zip(self._construct_line_string_expected_cords(10000), long_poly_obj.exterior.coords):
        #    self.assertEqual(expected, recieved)
        self.assertEqual(long_poly_obj.exterior.coords, self._construct_line_string_expected_cords(10000))

        # Test bad polygon strings
        bps = "POLYGONE ((30 10, 40 40, 20 40, 10 20, 30 10))"
        with self.assertRaises(ValueError):
            bpo = Polygon.from_wkt(bps)
        bps = "POLYGON (30 10, 40 40, 20 40, 10 20, 30 10)"
        with self.assertRaises(ValueError):
            bpo = Polygon.from_wkt(bps)

        # Polygons get truncated automatically
        ps = "POLYGON ((30 10, 40 40, 20, 10 20, 30))"
        po = Polygon.from_wkt(ps)
        expected_ex_coords = ((30.0, 10.0), (40.0, 40.0), (20.0,), (10.0, 20.0), (30.0,))
        self.assertEqual(po.exterior.coords, expected_ex_coords)

        # Test Polygon with NAN
        ps = "POLYGON ((NAN NAN, NAN NAN, NAN NAN, NAN, NAN))"
        po = Polygon.from_wkt(ps)
        for cords in po.exterior.coords:
            for cord in cords:
                self.assertTrue(math.isnan(cord))

    def _construct_line_string(self, num_of_points):
        # Constructs a arbitrarily long line string
        ls = "LINESTRING ("
        for i in range(0, num_of_points):
            ls += str(i)+" "+str(i)+', '
        return ls.rstrip(', ')+")"

    def _construct_line_string_expected_cords(self, num_of_points):
        # Constructs the corresponding expected coordinates for a linge string, and polygon string
        coords = []

        for i in range(0, num_of_points):
            coords.append((i, i))
        return tuple(coords)

    def _construct_polygon_string(self, num_of_points):
        # Constructs a arbitrarily long polygpn string
        ls = "POLYGON (("
        for i in range(0, num_of_points):
            ls += str(i)+" "+str(i)+', '
        return ls.rstrip(', ')+"))"



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from cassandra import ConsistencyLevel
from cassandra.query import SimpleStatement

from packaging.version import Version
from tests.integration import use_singledc, PROTOCOL_VERSION, \
    remove_cluster, greaterthanorequalcass40, notdse, \
    CASSANDRA_VERSION, DSE_VERSION, TestCluster, DEFAULT_SINGLE_INTERFACE_PORT


def setup_module():
    if not DSE_VERSION and CASSANDRA_VERSION >= Version('4-a'):
        remove_cluster()
        use_singledc(use_single_interface=True)

def teardown_module():
    remove_cluster()


@notdse
@greaterthanorequalcass40
class SingleInterfaceTest(unittest.TestCase):

    def setUp(self):
        self.cluster = TestCluster(port=DEFAULT_SINGLE_INTERFACE_PORT)
        self.session = self.cluster.connect()

    def tearDown(self):
        if self.cluster is not None:
            self.cluster.shutdown()

    def test_single_interface(self):
        """
        Test that we can connect to a multiple hosts bound to a single interface.
        """
        hosts = self.cluster.metadata._hosts
        broadcast_rpc_ports = []
        broadcast_ports = []
        self.assertEqual(len(hosts), 3)
        for endpoint, host in hosts.items():

            self.assertEqual(endpoint.address, host.broadcast_rpc_address)
            self.assertEqual(endpoint.port, host.broadcast_rpc_port)

            if host.broadcast_rpc_port in broadcast_rpc_ports:
                self.fail("Duplicate broadcast_rpc_port")
            broadcast_rpc_ports.append(host.broadcast_rpc_port)
            if host.broadcast_port in broadcast_ports:
                self.fail("Duplicate broadcast_port")
            broadcast_ports.append(host.broadcast_port)

        for _ in range(1, 100):
            self.session.execute(SimpleStatement("select * from system_distributed.view_build_status",
                                                 consistency_level=ConsistencyLevel.ALL))

        for pool in self.session.get_pools():
            self.assertEqual(1, pool.get_state()['open_count'])



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from io import BytesIO

from cassandra import DriverException
from cassandra.segment import Segment, CrcException
from cassandra.connection import segment_codec_no_compression, segment_codec_lz4


def to_bits(b):
    return '{:08b}'.format(b)

class SegmentCodecTest(unittest.TestCase):

    small_msg = b'b' * 50
    max_msg = b'b' * Segment.MAX_PAYLOAD_LENGTH
    large_msg = b'b' * (Segment.MAX_PAYLOAD_LENGTH + 1)

    @staticmethod
    def _header_to_bits(data):
        # unpack a header to bits
        # data should be the little endian bytes sequence
        if len(data) > 6:  # compressed
            data = data[:5]
            bits = ''.join([to_bits(b) for b in reversed(data)])
            # return the compressed payload length, the uncompressed payload length,
            # the self contained flag and the padding as bits
            return bits[23:40] + bits[6:23] + bits[5:6] + bits[:5]
        else:  # uncompressed
            data = data[:3]
            bits = ''.join([to_bits(b) for b in reversed(data)])
            # return the payload length, the self contained flag and
            # the padding as bits
            return bits[7:24] + bits[6:7] + bits[:6]

    def test_encode_uncompressed_header(self):
        buffer = BytesIO()
        segment_codec_no_compression.encode_header(buffer, len(self.small_msg), -1, True)
        self.assertEqual(buffer.tell(), 6)
        self.assertEqual(
            self._header_to_bits(buffer.getvalue()),
            "00000000000110010" + "1" + "000000")

    @unittest.skipUnless(segment_codec_lz4, ' lz4 not installed')
    def test_encode_compressed_header(self):
        buffer = BytesIO()
        compressed_length = len(segment_codec_lz4.compress(self.small_msg))
        segment_codec_lz4.encode_header(buffer, compressed_length, len(self.small_msg), True)

        self.assertEqual(buffer.tell(), 8)
        self.assertEqual(
            self._header_to_bits(buffer.getvalue()),
            "{:017b}".format(compressed_length) + "00000000000110010" + "1" + "00000")

    def test_encode_uncompressed_header_with_max_payload(self):
        buffer = BytesIO()
        segment_codec_no_compression.encode_header(buffer, len(self.max_msg), -1, True)
        self.assertEqual(buffer.tell(), 6)
        self.assertEqual(
            self._header_to_bits(buffer.getvalue()),
            "11111111111111111" + "1" + "000000")

    def test_encode_header_fails_if_payload_too_big(self):
        buffer = BytesIO()
        for codec in [c for c in [segment_codec_no_compression, segment_codec_lz4] if c is not None]:
            with self.assertRaises(DriverException):
                codec.encode_header(buffer, len(self.large_msg), -1, False)

    def test_encode_uncompressed_header_not_self_contained_msg(self):
        buffer = BytesIO()
        # simulate the first chunk with the max size
        segment_codec_no_compression.encode_header(buffer, len(self.max_msg), -1, False)
        self.assertEqual(buffer.tell(), 6)
        self.assertEqual(
            self._header_to_bits(buffer.getvalue()),
            ("11111111111111111"
             "0"  # not self contained
             "000000"))

    @unittest.skipUnless(segment_codec_lz4, ' lz4 not installed')
    def test_encode_compressed_header_with_max_payload(self):
        buffer = BytesIO()
        compressed_length = len(segment_codec_lz4.compress(self.max_msg))
        segment_codec_lz4.encode_header(buffer, compressed_length, len(self.max_msg), True)
        self.assertEqual(buffer.tell(), 8)
        self.assertEqual(
            self._header_to_bits(buffer.getvalue()),
            "{:017b}".format(compressed_length) + "11111111111111111" + "1" + "00000")

    @unittest.skipUnless(segment_codec_lz4, ' lz4 not installed')
    def test_encode_compressed_header_not_self_contained_msg(self):
        buffer = BytesIO()
        # simulate the first chunk with the max size
        compressed_length = len(segment_codec_lz4.compress(self.max_msg))
        segment_codec_lz4.encode_header(buffer, compressed_length, len(self.max_msg), False)
        self.assertEqual(buffer.tell(), 8)
        self.assertEqual(
            self._header_to_bits(buffer.getvalue()),
            ("{:017b}".format(compressed_length) +
             "11111111111111111"
             "0"  # not self contained
             "00000"))

    def test_decode_uncompressed_header(self):
        buffer = BytesIO()
        segment_codec_no_compression.encode_header(buffer, len(self.small_msg), -1, True)
        buffer.seek(0)
        header = segment_codec_no_compression.decode_header(buffer)
        self.assertEqual(header.uncompressed_payload_length, -1)
        self.assertEqual(header.payload_length, len(self.small_msg))
        self.assertEqual(header.is_self_contained, True)

    @unittest.skipUnless(segment_codec_lz4, ' lz4 not installed')
    def test_decode_compressed_header(self):
        buffer = BytesIO()
        compressed_length = len(segment_codec_lz4.compress(self.small_msg))
        segment_codec_lz4.encode_header(buffer, compressed_length, len(self.small_msg), True)
        buffer.seek(0)
        header = segment_codec_lz4.decode_header(buffer)
        self.assertEqual(header.uncompressed_payload_length, len(self.small_msg))
        self.assertEqual(header.payload_length, compressed_length)
        self.assertEqual(header.is_self_contained, True)

    def test_decode_header_fails_if_corrupted(self):
        buffer = BytesIO()
        segment_codec_no_compression.encode_header(buffer, len(self.small_msg), -1, True)
        # corrupt one byte
        buffer.seek(buffer.tell()-1)
        buffer.write(b'0')
        buffer.seek(0)

        with self.assertRaises(CrcException):
            segment_codec_no_compression.decode_header(buffer)

    def test_decode_uncompressed_self_contained_segment(self):
        buffer = BytesIO()
        segment_codec_no_compression.encode(buffer, self.small_msg)

        buffer.seek(0)
        header = segment_codec_no_compression.decode_header(buffer)
        segment = segment_codec_no_compression.decode(buffer, header)

        self.assertEqual(header.is_self_contained, True)
        self.assertEqual(header.uncompressed_payload_length, -1)
        self.assertEqual(header.payload_length, len(self.small_msg))
        self.assertEqual(segment.payload, self.small_msg)

    @unittest.skipUnless(segment_codec_lz4, ' lz4 not installed')
    def test_decode_compressed_self_contained_segment(self):
        buffer = BytesIO()
        segment_codec_lz4.encode(buffer, self.small_msg)

        buffer.seek(0)
        header = segment_codec_lz4.decode_header(buffer)
        segment = segment_codec_lz4.decode(buffer, header)

        self.assertEqual(header.is_self_contained, True)
        self.assertEqual(header.uncompressed_payload_length, len(self.small_msg))
        self.assertGreater(header.uncompressed_payload_length, header.payload_length)
        self.assertEqual(segment.payload, self.small_msg)

    def test_decode_multi_segments(self):
        buffer = BytesIO()
        segment_codec_no_compression.encode(buffer, self.large_msg)

        buffer.seek(0)
        # We should have 2 segments to read
        headers = []
        segments = []
        headers.append(segment_codec_no_compression.decode_header(buffer))
        segments.append(segment_codec_no_compression.decode(buffer, headers[0]))
        headers.append(segment_codec_no_compression.decode_header(buffer))
        segments.append(segment_codec_no_compression.decode(buffer, headers[1]))

        self.assertTrue(all([h.is_self_contained is False for h in headers]))
        decoded_msg = segments[0].payload + segments[1].payload
        self.assertEqual(decoded_msg, self.large_msg)

    @unittest.skipUnless(segment_codec_lz4, ' lz4 not installed')
    def test_decode_fails_if_corrupted(self):
        buffer = BytesIO()
        segment_codec_lz4.encode(buffer, self.small_msg)
        buffer.seek(buffer.tell()-1)
        buffer.write(b'0')
        buffer.seek(0)
        header = segment_codec_lz4.decode_header(buffer)
        with self.assertRaises(CrcException):
            segment_codec_lz4.decode(buffer, header)

    @unittest.skipUnless(segment_codec_lz4, ' lz4 not installed')
    def test_decode_tiny_msg_not_compressed(self):
        buffer = BytesIO()
        segment_codec_lz4.encode(buffer, b'b')
        buffer.seek(0)
        header = segment_codec_lz4.decode_header(buffer)
        segment = segment_codec_lz4.decode(buffer, header)
        self.assertEqual(header.uncompressed_payload_length, 0)
        self.assertEqual(header.payload_length, 1)
        self.assertEqual(segment.payload, b'b')



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from unittest.mock import patch

from cassandra.cqlengine import columns, CQLEngineException
from cassandra.cqlengine.management import sync_table, drop_table, create_keyspace_simple, drop_keyspace
from cassandra.cqlengine import models
from cassandra.cqlengine.models import Model, ModelDefinitionException
from uuid import uuid1
from tests.integration import pypy
from tests.integration.cqlengine.base import TestQueryUpdateModel

class TestModel(unittest.TestCase):
    """ Tests the non-io functionality of models """

    def test_instance_equality(self):
        """ tests the model equality functionality """
        class EqualityModel(Model):

            pk = columns.Integer(primary_key=True)

        m0 = EqualityModel(pk=0)
        m1 = EqualityModel(pk=1)

        self.assertEqual(m0, m0)
        self.assertNotEqual(m0, m1)

    def test_model_equality(self):
        """ tests the model equality functionality """
        class EqualityModel0(Model):

            pk = columns.Integer(primary_key=True)

        class EqualityModel1(Model):

            kk = columns.Integer(primary_key=True)

        m0 = EqualityModel0(pk=0)
        m1 = EqualityModel1(kk=1)

        self.assertEqual(m0, m0)
        self.assertNotEqual(m0, m1)

    def test_keywords_as_names(self):
        """
        Test for CQL keywords as names

        test_keywords_as_names tests that CQL keywords are properly and automatically quoted in cqlengine. It creates
        a keyspace, keyspace, which should be automatically quoted to "keyspace" in CQL. It then creates a table, table,
        which should also be automatically quoted to "table". It then verfies that operations can be done on the
        "keyspace"."table" which has been created. It also verifies that table alternations work and operations can be
        performed on the altered table.

        @since 2.6.0
        @jira_ticket PYTHON-244
        @expected_result Cqlengine should quote CQL keywords properly when creating keyspaces and tables.

        @test_category schema:generation
        """

        # If the keyspace exists, it will not be re-created
        create_keyspace_simple('keyspace', 1)

        class table(Model):
            __keyspace__ = 'keyspace'
            select = columns.Integer(primary_key=True)
            table = columns.Text()

        # In case the table already exists in keyspace
        drop_table(table)

        # Create should work
        sync_table(table)

        created = table.create(select=0, table='table')
        selected = table.objects(select=0)[0]
        self.assertEqual(created.select, selected.select)
        self.assertEqual(created.table, selected.table)

        # Alter should work
        class table(Model):
            __keyspace__ = 'keyspace'
            select = columns.Integer(primary_key=True)
            table = columns.Text()
            where = columns.Text()

        sync_table(table)

        created = table.create(select=1, table='table')
        selected = table.objects(select=1)[0]
        self.assertEqual(created.select, selected.select)
        self.assertEqual(created.table, selected.table)
        self.assertEqual(created.where, selected.where)

        drop_keyspace('keyspace')

    def test_column_family(self):
        class TestModel(Model):
            k = columns.Integer(primary_key=True)

        # no model keyspace uses default
        self.assertEqual(TestModel.column_family_name(), "%s.test_model" % (models.DEFAULT_KEYSPACE,))

        # model keyspace overrides
        TestModel.__keyspace__ = "my_test_keyspace"
        self.assertEqual(TestModel.column_family_name(), "%s.test_model" % (TestModel.__keyspace__,))

        # neither set should raise CQLEngineException before failing or formatting an invalid name
        del TestModel.__keyspace__
        with patch('cassandra.cqlengine.models.DEFAULT_KEYSPACE', None):
            self.assertRaises(CQLEngineException, TestModel.column_family_name)
            # .. but we can still get the bare CF name
            self.assertEqual(TestModel.column_family_name(include_keyspace=False), "test_model")

    def test_column_family_case_sensitive(self):
        """
        Test to ensure case sensitivity is honored when __table_name_case_sensitive__ flag is set

        @since 3.1
        @jira_ticket PYTHON-337
        @expected_result table_name case is respected

        @test_category object_mapper
        """
        class TestModel(Model):
            __table_name__ = 'TestModel'
            __table_name_case_sensitive__ = True

            k = columns.Integer(primary_key=True)

        self.assertEqual(TestModel.column_family_name(), '%s."TestModel"' % (models.DEFAULT_KEYSPACE,))

        TestModel.__keyspace__ = "my_test_keyspace"
        self.assertEqual(TestModel.column_family_name(), '%s."TestModel"' % (TestModel.__keyspace__,))

        del TestModel.__keyspace__
        with patch('cassandra.cqlengine.models.DEFAULT_KEYSPACE', None):
            self.assertRaises(CQLEngineException, TestModel.column_family_name)
            self.assertEqual(TestModel.column_family_name(include_keyspace=False), '"TestModel"')


class BuiltInAttributeConflictTest(unittest.TestCase):
    """tests Model definitions that conflict with built-in attributes/methods"""

    def test_model_with_attribute_name_conflict(self):
        """should raise exception when model defines column that conflicts with built-in attribute"""
        with self.assertRaises(ModelDefinitionException):
            class IllegalTimestampColumnModel(Model):

                my_primary_key = columns.Integer(primary_key=True)
                timestamp = columns.BigInt()

    def test_model_with_method_name_conflict(self):
        """should raise exception when model defines column that conflicts with built-in method"""
        with self.assertRaises(ModelDefinitionException):
            class IllegalFilterColumnModel(Model):

                my_primary_key = columns.Integer(primary_key=True)
                filter = columns.Text()


class ModelOverWriteTest(unittest.TestCase):

    def test_model_over_write(self):
        """
        Test to ensure overwriting of primary keys in model inheritance is allowed

        This is currently only an issue in PyPy. When PYTHON-504 is introduced this should
        be updated error out and warn the user

        @since 3.6.0
        @jira_ticket PYTHON-576
        @expected_result primary keys can be overwritten via inheritance

        @test_category object_mapper
        """
        class TimeModelBase(Model):
            uuid = columns.TimeUUID(primary_key=True)

        class DerivedTimeModel(TimeModelBase):
            __table_name__ = 'derived_time'
            uuid = columns.TimeUUID(primary_key=True, partition_key=True)
            value = columns.Text(required=False)

        # In case the table already exists in keyspace
        drop_table(DerivedTimeModel)

        sync_table(DerivedTimeModel)
        uuid_value = uuid1()
        uuid_value2 = uuid1()
        DerivedTimeModel.create(uuid=uuid_value, value="first")
        DerivedTimeModel.create(uuid=uuid_value2, value="second")
        DerivedTimeModel.objects.filter(uuid=uuid_value)


class TestColumnComparison(unittest.TestCase):
    def test_comparison(self):
        l = [TestQueryUpdateModel.partition.column,
             TestQueryUpdateModel.cluster.column,
             TestQueryUpdateModel.count.column,
             TestQueryUpdateModel.text.column,
             TestQueryUpdateModel.text_set.column,
             TestQueryUpdateModel.text_list.column,
             TestQueryUpdateModel.text_map.column]

        self.assertEqual(l, sorted(l))
        self.assertNotEqual(TestQueryUpdateModel.partition.column, TestQueryUpdateModel.cluster.column)
        self.assertLessEqual(TestQueryUpdateModel.partition.column, TestQueryUpdateModel.cluster.column)
        self.assertGreater(TestQueryUpdateModel.cluster.column, TestQueryUpdateModel.partition.column)
        self.assertGreaterEqual(TestQueryUpdateModel.cluster.column, TestQueryUpdateModel.partition.column)


class TestDeprecationWarning(unittest.TestCase):
    def test_deprecation_warnings(self):
        """
        Test to some deprecation warning have been added. It tests warnings for
        negative index, negative index slicing and table sensitive removal

        This test should be removed in 4.0, that's why the imports are in
        this test, so it's easier to remove

        @since 3.13
        @jira_ticket PYTHON-877
        @expected_result the deprecation warnings are emitted

        @test_category logs
        """
        import warnings

        class SensitiveModel(Model):
            __table_name__ = 'SensitiveModel'
            __table_name_case_sensitive__ = True
            k = columns.Integer(primary_key=True)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sync_table(SensitiveModel)
            self.addCleanup(drop_table, SensitiveModel)

            SensitiveModel.create(k=0)

            rows = SensitiveModel.objects().all().allow_filtering()
            rows[-1]
            rows[-1:]

            # Asyncio complains loudly about old syntax on python 3.7+, so get rid of all of those
            relevant_warnings = [warn for warn in w if "with (yield from lock)" not in str(warn.message)]

            self.assertEqual(len(relevant_warnings), 4)
            self.assertIn("__table_name_case_sensitive__ will be removed in 4.0.", str(relevant_warnings[0].message))
            self.assertIn("__table_name_case_sensitive__ will be removed in 4.0.", str(relevant_warnings[1].message))
            self.assertIn("ModelQuerySet indexing with negative indices support will be removed in 4.0.",
                          str(relevant_warnings[2].message))
            self.assertIn("ModelQuerySet slicing with negative indices support will be removed in 4.0.",
                          str(relevant_warnings[3].message))



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import time
import logging
from packaging.version import Version
from collections import namedtuple

from cassandra.cluster import EXEC_PROFILE_GRAPH_DEFAULT
from cassandra.graph import graph_result_row_factory
from cassandra.graph.query import GraphProtocol
from cassandra.graph.types import VertexProperty

from tests.util import wait_until
from tests.integration.advanced.graph import BasicGraphUnitTestCase, ClassicGraphFixtures, \
    ClassicGraphSchema, CoreGraphSchema
from tests.integration.advanced.graph import VertexLabel, GraphTestConfiguration, GraphUnitTestCase
from tests.integration import DSE_VERSION, requiredse

log = logging.getLogger(__name__)


@requiredse
class GraphBasicDataTypesTests(BasicGraphUnitTestCase):

    def test_result_types(self):
        """
        Test to validate that the edge and vertex version of results are constructed correctly.

        @since 1.0.0
        @jira_ticket PYTHON-479
        @expected_result edge/vertex result types should be unpacked correctly.
        @test_category dse graph
        """
        queries, params = ClassicGraphFixtures.multiple_fields()
        for query in queries:
            self.session.execute_graph(query, params)

        prof = self.session.execution_profile_clone_update(EXEC_PROFILE_GRAPH_DEFAULT, row_factory=graph_result_row_factory)  # requires simplified row factory to avoid shedding id/~type information used for validation below
        rs = self.session.execute_graph("g.V()", execution_profile=prof)

        for result in rs:
            self._validate_type(result)

    def _validate_type(self, vertex):
        for properties in vertex.properties.values():
            prop = properties[0]

            if DSE_VERSION >= Version("5.1"):
                type_indicator = prop['id']['~label']
            else:
                type_indicator = prop['id']['~type']

            if any(type_indicator.startswith(t) for t in
                   ('int', 'short', 'long', 'bigint', 'decimal', 'smallint', 'varint')):
                typ = int
            elif any(type_indicator.startswith(t) for t in ('float', 'double')):
                typ = float
            elif any(type_indicator.startswith(t) for t in ('duration', 'date', 'negdate', 'time',
                                                            'blob', 'timestamp', 'point', 'linestring', 'polygon',
                                                            'inet', 'uuid')):
                typ = str
            else:
                pass
                self.fail("Received unexpected type: %s" % type_indicator)
            self.assertIsInstance(prop['value'], typ)


class GenericGraphDataTypeTest(GraphUnitTestCase):

    def _test_all_datatypes(self, schema, graphson):
        ep = self.get_execution_profile(graphson)

        for data in schema.fixtures.datatypes().values():
            typ, value, deserializer = data
            vertex_label = VertexLabel([typ])
            property_name = next(iter(vertex_label.non_pk_properties.keys()))
            schema.create_vertex_label(self.session, vertex_label, execution_profile=ep)
            vertex = list(schema.add_vertex(self.session, vertex_label, property_name, value, execution_profile=ep))[0]

            def get_vertex_properties():
                return list(schema.get_vertex_properties(
                    self.session, vertex, execution_profile=ep))

            prop_returned = 1 if DSE_VERSION < Version('5.1') else 2  # include pkid >=5.1
            wait_until(
                lambda: len(get_vertex_properties()) == prop_returned, 0.2, 15)

            vertex_properties = get_vertex_properties()
            if graphson == GraphProtocol.GRAPHSON_1_0:
                vertex_properties = [vp.as_vertex_property() for vp in vertex_properties]

            for vp in vertex_properties:
                if vp.label == 'pkid':
                    continue

                self.assertIsInstance(vp, VertexProperty)
                self.assertEqual(vp.label, property_name)
                if graphson == GraphProtocol.GRAPHSON_1_0:
                    deserialized_value = deserializer(vp.value) if deserializer else vp.value
                    self.assertEqual(deserialized_value, value)
                else:
                    self.assertEqual(vp.value, value)

    def __test_udt(self, schema, graphson, address_class, address_with_tags_class,
                   complex_address_class, complex_address_with_owners_class):
        if schema is not CoreGraphSchema or DSE_VERSION < Version('6.8'):
            raise unittest.SkipTest("Graph UDT is only supported with DSE 6.8+ and Core graphs.")

        ep = self.get_execution_profile(graphson)

        Address = address_class
        AddressWithTags = address_with_tags_class
        ComplexAddress = complex_address_class
        ComplexAddressWithOwners = complex_address_with_owners_class

        # setup udt
        self.session.execute_graph("""
                schema.type('address').property('address', Text).property('city', Text).property('state', Text).create();
                schema.type('addressTags').property('address', Text).property('city', Text).property('state', Text).
                    property('tags', setOf(Text)).create();
                schema.type('complexAddress').property('address', Text).property('address_tags', frozen(typeOf('addressTags'))).
                    property('city', Text).property('state', Text).property('props', mapOf(Text, Int)).create();
                schema.type('complexAddressWithOwners').property('address', Text).
                    property('address_tags', frozen(typeOf('addressTags'))).
                    property('city', Text).property('state', Text).property('props', mapOf(Text, Int)).
                    property('owners', frozen(listOf(tupleOf(Text, Int)))).create();
                """, execution_profile=ep)

        time.sleep(2)  # wait the UDT to be discovered
        self.session.cluster.register_user_type(self.graph_name, 'address', Address)
        self.session.cluster.register_user_type(self.graph_name, 'addressTags', AddressWithTags)
        self.session.cluster.register_user_type(self.graph_name, 'complexAddress', ComplexAddress)
        self.session.cluster.register_user_type(self.graph_name, 'complexAddressWithOwners', ComplexAddressWithOwners)

        data = {
            "udt1": ["typeOf('address')", Address('1440 Rd Smith', 'Quebec', 'QC')],
            "udt2": ["tupleOf(typeOf('address'), Text)", (Address('1440 Rd Smith', 'Quebec', 'QC'), 'hello')],
            "udt3": ["tupleOf(frozen(typeOf('address')), Text)", (Address('1440 Rd Smith', 'Quebec', 'QC'), 'hello')],
            "udt4": ["tupleOf(tupleOf(Int, typeOf('address')), Text)",
                      ((42, Address('1440 Rd Smith', 'Quebec', 'QC')), 'hello')],
            "udt5": ["tupleOf(tupleOf(Int, typeOf('addressTags')), Text)",
                     ((42, AddressWithTags('1440 Rd Smith', 'Quebec', 'QC', {'t1', 't2'})), 'hello')],
            "udt6": ["tupleOf(tupleOf(Int, typeOf('complexAddress')), Text)",
                     ((42, ComplexAddress('1440 Rd Smith',
                                          AddressWithTags('1440 Rd Smith', 'Quebec', 'QC', {'t1', 't2'}),
                                          'Quebec', 'QC', {'p1': 42, 'p2': 33})), 'hello')],
            "udt7": ["tupleOf(tupleOf(Int, frozen(typeOf('complexAddressWithOwners'))), Text)",
                     ((42, ComplexAddressWithOwners(
                           '1440 Rd Smith',
                           AddressWithTags('1440 CRd Smith', 'Quebec', 'QC', {'t1', 't2'}),
                           'Quebec', 'QC', {'p1': 42, 'p2': 33}, [('Mike', 43), ('Gina', 39)])
                       ), 'hello')]
        }

        for typ, value in data.values():
            vertex_label = VertexLabel([typ])
            property_name = next(iter(vertex_label.non_pk_properties.keys()))
            schema.create_vertex_label(self.session, vertex_label, execution_profile=ep)

            vertex = list(schema.add_vertex(self.session, vertex_label, property_name, value, execution_profile=ep))[0]

            def get_vertex_properties():
                return list(schema.get_vertex_properties(
                    self.session, vertex, execution_profile=ep))

            wait_until(
                lambda: len(get_vertex_properties()) == 2, 0.2, 15)

            vertex_properties = get_vertex_properties()
            for vp in vertex_properties:
                if vp.label == 'pkid':
                    continue

                self.assertIsInstance(vp, VertexProperty)
                self.assertEqual(vp.label, property_name)
                self.assertEqual(vp.value, value)

    def _test_udt_with_classes(self, schema, graphson):
        class Address(object):

            def __init__(self, address, city, state):
                self.address = address
                self.city = city
                self.state = state

            def __eq__(self, other):
                return self.address == other.address and self.city == other.city and self.state == other.state

        class AddressWithTags(object):

            def __init__(self, address, city, state, tags):
                self.address = address
                self.city = city
                self.state = state
                self.tags = tags

            def __eq__(self, other):
                return (self.address == other.address and self.city == other.city
                        and self.state == other.state and self.tags == other.tags)

        class ComplexAddress(object):

            def __init__(self, address, address_tags, city, state, props):
                self.address = address
                self.address_tags = address_tags
                self.city = city
                self.state = state
                self.props = props

            def __eq__(self, other):
                return (self.address == other.address and self.address_tags == other.address_tags
                        and self.city == other.city and self.state == other.state
                        and self.props == other.props)

        class ComplexAddressWithOwners(object):

            def __init__(self, address, address_tags, city, state, props, owners):
                self.address = address
                self.address_tags = address_tags
                self.city = city
                self.state = state
                self.props = props
                self.owners = owners

            def __eq__(self, other):
                return (self.address == other.address and self.address_tags == other.address_tags
                        and self.city == other.city and self.state == other.state
                        and self.props == other.props and self.owners == other.owners)

        self.__test_udt(schema, graphson, Address, AddressWithTags, ComplexAddress, ComplexAddressWithOwners)

    def _test_udt_with_namedtuples(self, schema, graphson):
        AddressTuple = namedtuple('Address', ('address', 'city', 'state'))
        AddressWithTagsTuple = namedtuple('AddressWithTags', ('address', 'city', 'state', 'tags'))
        ComplexAddressTuple = namedtuple('ComplexAddress', ('address', 'address_tags', 'city', 'state', 'props'))
        ComplexAddressWithOwnersTuple = namedtuple('ComplexAddressWithOwners', ('address', 'address_tags', 'city',
                                                                                'state', 'props', 'owners'))

        self.__test_udt(schema, graphson, AddressTuple, AddressWithTagsTuple,
                        ComplexAddressTuple, ComplexAddressWithOwnersTuple)


@requiredse
@GraphTestConfiguration.generate_tests(schema=ClassicGraphSchema)
class ClassicGraphDataTypeTest(GenericGraphDataTypeTest):
    pass


@requiredse
@GraphTestConfiguration.generate_tests(schema=CoreGraphSchema)
class CoreGraphDataTypeTest(GenericGraphDataTypeTest):
    pass



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from tests.integration import use_singledc, TestCluster

from cassandra.policies import ColDesc

from cassandra.column_encryption.policies import AES256ColumnEncryptionPolicy, \
    AES256_KEY_SIZE_BYTES, AES256_BLOCK_SIZE_BYTES

def setup_module():
    use_singledc()

class ColumnEncryptionPolicyTest(unittest.TestCase):

    def _recreate_keyspace(self, session):
        session.execute("drop keyspace if exists foo")
        session.execute("CREATE KEYSPACE foo WITH replication = {'class': 'SimpleStrategy', 'replication_factor': '1'}")
        session.execute("CREATE TABLE foo.bar(encrypted blob, unencrypted int, primary key(unencrypted))")

    def _create_policy(self, key, iv = None):
        cl_policy = AES256ColumnEncryptionPolicy()
        col_desc = ColDesc('foo','bar','encrypted')
        cl_policy.add_column(col_desc, key, "int")
        return (col_desc, cl_policy)

    def test_end_to_end_prepared(self):

        # We only currently perform testing on a single type/expected value pair since CLE functionality is essentially
        # independent of the underlying type.  We intercept data after it's been encoded when it's going out and before it's
        # encoded when coming back; the actual types of the data involved don't impact us.
        expected = 0

        key = os.urandom(AES256_KEY_SIZE_BYTES)
        (_, cl_policy) = self._create_policy(key)
        cluster = TestCluster(column_encryption_policy=cl_policy)
        session = cluster.connect()
        self._recreate_keyspace(session)

        prepared = session.prepare("insert into foo.bar (encrypted, unencrypted) values (?,?)")
        for i in range(100):
            session.execute(prepared, (i, i))

        # A straight select from the database will now return the decrypted bits.  We select both encrypted and unencrypted
        # values here to confirm that we don't interfere with regular processing of unencrypted vals.
        (encrypted,unencrypted) = session.execute("select encrypted, unencrypted from foo.bar where unencrypted = %s allow filtering", (expected,)).one()
        self.assertEqual(expected, encrypted)
        self.assertEqual(expected, unencrypted)

        # Confirm the same behaviour from a subsequent prepared statement as well
        prepared = session.prepare("select encrypted, unencrypted from foo.bar where unencrypted = ? allow filtering")
        (encrypted,unencrypted) = session.execute(prepared, [expected]).one()
        self.assertEqual(expected, encrypted)
        self.assertEqual(expected, unencrypted)

    def test_end_to_end_simple(self):

        expected = 1

        key = os.urandom(AES256_KEY_SIZE_BYTES)
        (col_desc, cl_policy) = self._create_policy(key)
        cluster = TestCluster(column_encryption_policy=cl_policy)
        session = cluster.connect()
        self._recreate_keyspace(session)

        # Use encode_and_encrypt helper function to populate date
        for i in range(1,100):
            self.assertIsNotNone(i)
            encrypted = cl_policy.encode_and_encrypt(col_desc, i)
            session.execute("insert into foo.bar (encrypted, unencrypted) values (%s,%s)", (encrypted, i))

        # A straight select from the database will now return the decrypted bits.  We select both encrypted and unencrypted
        # values here to confirm that we don't interfere with regular processing of unencrypted vals.
        (encrypted,unencrypted) = session.execute("select encrypted, unencrypted from foo.bar where unencrypted = %s allow filtering", (expected,)).one()
        self.assertEqual(expected, encrypted)
        self.assertEqual(expected, unencrypted)

        # Confirm the same behaviour from a subsequent prepared statement as well
        prepared = session.prepare("select encrypted, unencrypted from foo.bar where unencrypted = ? allow filtering")
        (encrypted,unencrypted) = session.execute(prepared, [expected]).one()
        self.assertEqual(expected, encrypted)
        self.assertEqual(expected, unencrypted)

    def test_end_to_end_different_cle_contexts_different_ivs(self):
        """
        Test to validate PYTHON-1350.  We should be able to decode the data from two different contexts (with two different IVs)
        since the IV used to decrypt the data is actually now stored with the data.
        """

        expected = 2

        key = os.urandom(AES256_KEY_SIZE_BYTES)

        # Simulate the creation of two AES256 policies at two different times.  Python caches
        # default param args at function definition time so a single value will be used any time
        # the default val is used.  Upshot is that within the same test we'll always have the same
        # IV if we rely on the default args, so manually introduce some variation here to simulate
        # what actually happens if you have two distinct sessions created at two different times.
        iv1 = os.urandom(AES256_BLOCK_SIZE_BYTES)
        (col_desc1, cl_policy1) = self._create_policy(key, iv=iv1)
        cluster1 = TestCluster(column_encryption_policy=cl_policy1)
        session1 = cluster1.connect()
        self._recreate_keyspace(session1)

        # Use encode_and_encrypt helper function to populate date
        for i in range(1,100):
            self.assertIsNotNone(i)
            encrypted = cl_policy1.encode_and_encrypt(col_desc1, i)
            session1.execute("insert into foo.bar (encrypted, unencrypted) values (%s,%s)", (encrypted, i))
        session1.shutdown()
        cluster1.shutdown()

        # Explicitly clear the class-level cache here; we're trying to simulate a second connection from a completely new process and
        # that would entail not re-using any cached ciphers
        AES256ColumnEncryptionPolicy._build_cipher.cache_clear()
        cache_info = cl_policy1.cache_info()
        self.assertEqual(cache_info.currsize, 0)

        iv2 = os.urandom(AES256_BLOCK_SIZE_BYTES)
        (_, cl_policy2) = self._create_policy(key, iv=iv2)
        cluster2 = TestCluster(column_encryption_policy=cl_policy2)
        session2 = cluster2.connect()
        (encrypted,unencrypted) = session2.execute("select encrypted, unencrypted from foo.bar where unencrypted = %s allow filtering", (expected,)).one()
        self.assertEqual(expected, encrypted)
        self.assertEqual(expected, unencrypted)

    def test_end_to_end_different_cle_contexts_different_policies(self):
        """
        Test to validate PYTHON-1356.  Class variables used to pass CLE policy down to protocol handler shouldn't persist.
        """

        expected = 3

        key = os.urandom(AES256_KEY_SIZE_BYTES)
        (col_desc, cl_policy) = self._create_policy(key)
        cluster = TestCluster(column_encryption_policy=cl_policy)
        session = cluster.connect()
        self._recreate_keyspace(session)

        # Use encode_and_encrypt helper function to populate date
        session.execute("insert into foo.bar (encrypted, unencrypted) values (%s,%s)",(cl_policy.encode_and_encrypt(col_desc, expected), expected))

        # We now open a new session _without_ the CLE policy specified.  We should _not_ be able to read decrypted bits from this session.
        cluster2 = TestCluster()
        session2 = cluster2.connect()

        # A straight select from the database will now return the decrypted bits.  We select both encrypted and unencrypted
        # values here to confirm that we don't interfere with regular processing of unencrypted vals.
        (encrypted,unencrypted) = session2.execute("select encrypted, unencrypted from foo.bar where unencrypted = %s allow filtering", (expected,)).one()
        self.assertEqual(cl_policy.encode_and_encrypt(col_desc, expected), encrypted)
        self.assertEqual(expected, unencrypted)

        # Confirm the same behaviour from a subsequent prepared statement as well
        prepared = session2.prepare("select encrypted, unencrypted from foo.bar where unencrypted = ? allow filtering")
        (encrypted,unencrypted) = session2.execute(prepared, [expected]).one()
        self.assertEqual(cl_policy.encode_and_encrypt(col_desc, expected), encrypted)
